"""This module contains the state information of the heat demand model."""

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class HeatDemandState(ModelState):
    """Captures the state of the heat demand model"""

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.t_last_3_deg_celsius = inits.get("t_last_3_deg_celsius", 10.3)
        self.t_last_2_deg_celsius = inits.get("t_last_2_deg_celsius", 10.3)
        self.t_last_1_deg_celsius = inits.get("t_last_1_deg_celsius", 10.3)

        self.e_th_kwh: float = 0.0



"""This module contains the *mosaik_api* meta definition for pysimmods
models.
"""

from pysimmods.buffer import Battery
from pysimmods.consumer import HVAC
from pysimmods.generator import (
    BiogasPlant,
    CHPLPGSystem,
    PVPlantSystem,
    WindPowerPlantSystem,
)
from pysimmods.other.dummy.generator import DummyGenerator as DieselGenerator

_SHARED_ATTRS = [
    "p_set_mw",
    "p_set_kw",
    "p_mw",
    "p_kw",
    "q_mvar",
    "q_kvar",
    "local_time",
]


META = {
    "type": "time-based",
    "models": {
        "Battery": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["soc_percent"] + _SHARED_ATTRS,
        },
        "Photovoltaic": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "q_set_mvar",
                "q_set_kvar",
                "t_air_deg_celsius",
                "t_module_deg_celsius",
                "bh_w_per_m2",
                "dh_w_per_m2",
                "s_module_w_per_m2",
                "inverter_inductive",
                "p_possible_max_mw",
            ]
            + _SHARED_ATTRS,
        },
        "CHP": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "day_avg_t_air_deg_celsius",
                "p_th_mw",
                "e_th_demand_set_kwh",
                "e_th_demand_kwh",
            ]
            + _SHARED_ATTRS,
        },
        "HVAC": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["t_air_deg_celsius", "theta_t_deg_celsius"]
            + _SHARED_ATTRS,
        },
        "Biogas": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["p_th_mw"] + _SHARED_ATTRS,
        },
        "DieselGenerator": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": _SHARED_ATTRS,
        },
        "WindTurbine": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "q_set_mvar",
                "q_set_kvar",
                "wind_v_m_per_s",
                "t_air_deg_celsius",
                "air_pressure_hpa",
                "inverter_inductive",
                "p_possible_max_mw",
            ]
            + _SHARED_ATTRS,
        },
    },
}

MODELS = {
    "Photovoltaic": PVPlantSystem,
    "HVAC": HVAC,
    "Battery": Battery,
    "CHP": CHPLPGSystem,
    "DieselGenerator": DieselGenerator,
    "Biogas": BiogasPlant,
    "WindTurbine": WindPowerPlantSystem,
}



"""This module contains a :class:`mosaik_api.Simulator` for the
flexiblity model, which is a wrapper for all models of the
pysimmods package.

"""

from __future__ import annotations

import json
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Union, cast

import mosaik_api_v3
import numpy as np
import pandas as pd
from midas.util.logging import set_and_init_logger
from mosaik_api_v3.types import (
    CreateResult,
    InputData,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)
from typing_extensions import override

from pysimmods.base.model import Model
from pysimmods.mosaik_bridge import LOG
from pysimmods.mosaik_bridge.meta import MODELS
from pysimmods.mosaik_bridge.pysim_mosaik import PysimmodsSimulator
from pysimmods.other.flexibility.flexibility_model import FlexibilityModel
from pysimmods.other.flexibility.forecast_model import ForecastModel
from pysimmods.other.flexibility.schedule_model import ScheduleModel
from pysimmods.util.date_util import GER


class FlexibilitySimulator(PysimmodsSimulator):
    """The generic flexiblity mosaik simulator for all pysimmods."""

    def __init__(self):
        super().__init__()

        self.models: dict[str, Model] = {}
        self.num_models: dict[ModelName, int] = {}

        self.sid: SimId = ""
        self.step_size: int = 0
        self.now_dt: datetime = datetime(1970, 1, 1)

        self.unit: str = "kw"
        self.use_decimal_percent: bool = False
        self.prioritize_setpoint: bool = False
        self.provide_forecasts: bool = False
        self.provide_flexibilities: bool = False
        self.forecast_horizon_hours: float = 0.25
        self.planning_horizon_hours: float = 1.0
        self.flexibility_horizon_hours: float = 2.0
        self.num_schedules: int = 10
        self.rng: np.random.Generator = np.random.default_rng()

    @override
    def init(
        self,
        sid: SimId,
        time_resolution: float = 1.0,
        start_date: str = "2010-01-01 00:00:00+0100",
        step_size: int = 900,
        unit: str = "kw",
        use_decimal_percent: bool = False,
        prioritize_setpoint: bool = False,
        provide_forecasts: bool = False,
        forecast_horizon_hours: float = 0.25,
        provide_flexibilities: bool = False,
        planning_horizon_hours: float = 1.0,
        flexibility_horizon_hours: float = 2.0,
        num_schedules: int = 10,
        seed: int | None = None,
        **sim_params: Dict[str, Any],
    ) -> Meta:
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.
        unit: str

        use_decimal_percent

        prioritize_setpoint

        provide_forecasts

        forecast_horizon_hours

        provide_flexibilities

        planning_horizon_hours

        flexibility_horizon_hours

        num_schedules

        seed

        key_value_logs

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        self.step_size = step_size
        self.now_dt = datetime.strptime(start_date, GER).astimezone(
            timezone.utc
        )

        self.unit = unit
        self.use_decimal_percent = use_decimal_percent

        self.prioritize_setpoint = prioritize_setpoint
        self.provide_forecasts = provide_forecasts
        self.forecast_horizon_hours = forecast_horizon_hours

        self.provide_flexibilities = provide_flexibilities
        self.planning_horizon_hours = planning_horizon_hours
        self.flexibility_horizon_hours = flexibility_horizon_hours
        self.num_schedules = num_schedules
        self.rng = np.random.default_rng(seed)

        self._update_meta()

        return self.meta

    @override
    def create(
        self,
        num: int,
        model: ModelName,
        *,
        params: Dict[str, Any],
        inits: Dict[str, Any],
        **model_params,  # for compatibility
    ) -> list[CreateResult]:
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.
        params: Dict[str, Any]
            The parameters dictionary for the model to create.
        inits: Dict[str, Any]
            The initial state dictionary for the model to create.

        Returns
        -------
        List[Dict[str, str]]
            A list with information on the created entity.

        """
        entities: list[CreateResult] = []
        params.setdefault("use_decimal_percent", self.use_decimal_percent)
        self.num_models.setdefault(model, 0)

        for _ in range(num):
            eid = f"{model}-{self.num_models[model]}"

            if self.provide_flexibilities:
                self.models[eid] = FlexibilityModel(
                    MODELS[model](params, inits),
                    unit=self.unit,
                    prioritize_setpoint=self.prioritize_setpoint,
                    forecast_horizon_hours=self.forecast_horizon_hours,
                    seed=self.rng.integers(2**32 - 1),
                )

            elif self.provide_forecasts:
                self.models[eid] = ForecastModel(
                    MODELS[model](params, inits),
                    unit=self.unit,
                    prioritize_setpoint=self.prioritize_setpoint,
                    forecast_horizon_hours=self.forecast_horizon_hours,
                )

            else:
                self.models[eid] = ScheduleModel(
                    MODELS[model](params, inits),
                    unit=self.unit,
                    prioritize_setpoint=self.prioritize_setpoint,
                )

            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})
        return entities

    @override
    def step(
        self, time: Time, inputs: InputData, max_advance: Time = 0
    ) -> Time | None:
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """

        self._set_default_inputs()

        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():
                if "forecast" in attr:
                    self._set_attr_forecast(eid, src_ids)
                elif attr == "schedule":
                    self._set_attr_schedule(eid, src_ids)
                elif attr == "local_time":
                    self._set_attr_local_time(eid, src_ids)
                else:
                    attr_sum = self._aggregate_attr(src_ids)
                    self._set_remaining_attrs(eid, attr, attr_sum)

        for model in self.models.values():
            model.step()

        self.now_dt += timedelta(seconds=self.step_size)
        self._generate_flexibilities()

        return time + self.step_size

    @override
    def get_data(self, outputs: OutputRequest) -> OutputData:
        """Returns the requested outputs (if feasible)"""

        data: Dict[str, Dict[str, Any]] = {}
        for eid, attrs in outputs.items():
            data[eid] = {}

            log_msg: dict[str, str | float | int | None] = {
                "id": f"{self.sid}_{eid}",
                "name": eid,
                "type": eid.split("-")[0],
            }

            for attr in attrs:
                if attr == "flexibilities":
                    value = self._get_attr_flexibilities(eid)
                    log_msg[attr] = json.loads(value)

                elif attr == "schedule":
                    value = self._get_attr_schedule(eid)
                    log_msg[attr] = json.loads(value)

                else:
                    value = self._get_remaining_attrs(eid, attr)
                    log_msg[attr] = value

                data.setdefault(eid, dict())[attr] = value

            LOG.debug(json.dumps(log_msg))

        return data

    def _update_meta(self):
        for model in self.meta["models"].keys():
            self.meta["models"][model]["attrs"].extend(
                ["flexibilities", "schedule", "target"]
            )

        self.meta["models"]["Photovoltaic"]["attrs"].extend(
            [
                "forecast_t_air_deg_celsius",
                "forecast_bh_w_per_m2",
                "forecast_dh_w_per_m2",
            ]
        )
        self.meta["models"]["CHP"]["attrs"].extend(
            ["forecast_day_avg_t_air_deg_celsius"]
        )
        self.meta["models"]["HVAC"]["attrs"].extend(
            ["forecast_t_air_deg_celsius"]
        )

    def _set_attr_forecast(self, eid: str, src_ids: Dict[str, Any]):
        for forecast in src_ids.values():
            if not isinstance(forecast, pd.DataFrame):
                forecast = pd.read_json(forecast).tz_localize("UTC")
            cast("ForecastModel", self.models[eid]).update_forecasts(forecast)

    def _set_attr_schedule(self, eid: str, src_ids: Dict[str, Any]):
        for schedule in src_ids.values():
            if schedule is not None:
                schedule = deserialize_schedule(schedule)

                if not schedule.empty:
                    cast("ScheduleModel", self.models[eid]).update_schedule(
                        schedule
                    )

    def _get_attr_schedule(self, eid: str) -> str:
        value = cast("ScheduleModel", self.models[eid]).schedule.to_json(
            self.now_dt,
            self.now_dt
            + timedelta(hours=self.forecast_horizon_hours)
            - timedelta(seconds=self.step_size),
        )

        return value

    def _get_attr_flexibilities(self, eid: str) -> str:
        dict_of_json = cast(
            "FlexibilityModel", self.models[eid]
        ).flexibilities.to_json()
        # value = json.dumps(dict_of_json)
        return dict_of_json

    def _generate_flexibilities(self):
        if self.provide_flexibilities:
            for model in self.models.values():
                cast("FlexibilityModel", model).generate_schedules(
                    (
                        self.now_dt
                        + timedelta(hours=self.planning_horizon_hours)
                    ).strftime(GER),
                    self.flexibility_horizon_hours,
                    self.num_schedules,
                )


def deserialize_schedule(
    schedule: Union[pd.DataFrame, Dict[str, Any], str],
) -> pd.DataFrame:
    """Convert the schedule provided by mosaik to DataFrame"""

    if isinstance(schedule, pd.DataFrame):
        return schedule

    if isinstance(schedule, dict):
        # The schedule might be nested into the schedule because of an
        # ICT simulator
        return deserialize_schedule(list(schedule.values())[0])

    if isinstance(schedule, str):
        return pd.read_json(schedule).tz_localize("UTC")

    raise ValueError(
        f"Unsupported schedule format {type(schedule)}: {schedule}"
    )


if __name__ == "__main__":
    set_and_init_logger(
        0, "pysimmods-logfile", "pysimmods-flex.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api_v3.start_simulation(FlexibilitySimulator())



from __future__ import annotations

import json

from pysimmods.other.flexibility.schedule import Schedule


class Flexibilities:
    def __init__(self):
        self._schedules: dict[int, Schedule] = {}

    def add_schedule(self, schedule_id: int, schedule: Schedule):
        self._schedules[int(schedule_id)] = schedule

    def __len__(self):
        return len(self._schedules)

    def items(self):
        for key, val in self._schedules.items():
            yield key, val

    def to_json(self):
        d = {key: val.to_dict() for key, val in self._schedules.items()}

        return json.dumps(d)

    def from_json(self, json_str: str) -> Flexibilities:
        self._schedules = {}
        d = json.loads(json_str)

        for key, schedule_dict in d.items():
            self._schedules[int(key)] = Schedule().from_dict(schedule_dict)

        return self

    def __getitem__(self, key: int):
        return self._schedules[key]



"""This module contains the state model for HVAC."""

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class HVACState(ModelState):
    """State parameters of the HVAC model.

    See :class:`.ModelState` for additional information.

    Parameters
    ----------
    inits : dict
        Contains the initial configuration of this HVAC model. See
        the attributes section for specific information of the HVAC
        model.

    Attributes
    ----------
    mass_kg : float
        Mass of the content in [kg].
    c_j_per_kg_k : float
        Mean specific heat capacity of refer cargo in [Jkg^-K^-1].
    theta_t_deg_celsius : float
        Time dependent temperature of the model in [°C].
    cooling : bool
        If set to True, the model is cooling, i.e., consuming power.


    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.mass_kg: float = inits.get("mass_kg", 500)
        self.c_j_per_kg_k: float = inits.get("c_j_per_kg_k", 2390)
        self.theta_t_deg_celsius: float = inits.get("theta_t_deg_celsius", 21)
        self.cooling: bool = inits.get("cooling", True)
        if isinstance(self.cooling, str):
            self.cooling = self.cooling.lower() not in ("0", "f", "false")



def linear_gradient(
    t_air_deg_kelvin: float,
    temperature_height_m: float,
    hub_height_m: float,
    gradient: float = 0.0065,
) -> float:
    """Calculate temperature at altitude h using linear gradient.
    
    Parameters
    ----------
    t_air_deg_kelvin: float
        Measured air temperature in [°K]
    temperature_height_m: float
        Height on which temperature was measured in [m].
    hub_height_m: float
        Height for which temperature should be calculated in [m].
    gradient: float
        The gradient used for the calculation.
    
    Returns
    -------
    float
        The temperature at altitude `hub_height_m` in [°K].
    
    """
    return t_air_deg_kelvin - gradient * (hub_height_m - temperature_height_m)



"""This module contains the config model for the CHP LPG."""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class CHPLPGConfig(ModelConfig):
    """Config parameters of the CHP LPG.

    This class capturs the configuration parameters of the chp model.

    Parameters
    ----------
    params : dict
        A dictionary containing the configuration parameters.

    Attributes
    ----------
    pn_min_kw : float
        Minimal nominal electrical power output in [kW].
    pn_max_kw : float
        Maximal nominal electrical power output in [kW].
    p_2_p_th_percent : float
        Ratio to compute thermal output from electrical power in [%].
    eta_min_percent : float
        Minimal total efficiency regarding electrical and thermal power
        in [%].
    eta_max_percent : float
        Maximal total efficiency regarding electrical and thermal power
        in [%].
    own_consumption_kw : float
        Own electrical power consumption of the unit [kW].
    active_min_s : int
        Minimal active time of the unit in [s].
    inactive_min_s : int
        Minimal inactive time of the unit in [s].
    lubricant_max_l : float
        Capacity of the lubricant storage in [l].
    lubricant_ml_per_h : float
        Consumption of lubricant per hour in [ml/h].

    """

    def __init__(self, params: ModelParams) -> None:
        super().__init__(params)

        self.p_min_kw = abs(cast("float", params["p_min_kw"]))
        self.p_max_kw = abs(cast("float", params["p_max_kw"]))

        self.p_2_p_th_percent = cast("float", params["p_2_p_th_percent"])
        self.eta_min_percent = cast("float", params["eta_min_percent"])
        self.eta_max_percent = cast("float", params["eta_max_percent"])
        self.own_consumption_kw = cast("float", params["own_consumption_kw"])
        self.active_min_s = cast("int", params["active_min_s"])
        self.inactive_min_s = cast("int", params["inactive_min_s"])

        self.lubricant_max_l = cast("float", params["lubricant_max_l"])
        self.lubricant_ml_per_h = cast("float", params["lubricant_ml_per_h"])

        self.default_schedule: list[float] = (
            [50.0, 50.0, 50.0, 50.0, 50.0, 100.0]
            + [100.0, 100.0, 100.0, 50.0, 50.0, 50.0]
            + [100.0, 100.0, 50.0, 50.0, 50.0, 100.0]
            + [100.0, 100.0, 100.0, 50.0, 50.0, 50.0]
        )



"""This module contains a Wind turbine."""

from copy import deepcopy

import numpy as np

from pysimmods.base.generator import Generator
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.generator.windsim.config import WindPowerPlantConfig
from pysimmods.generator.windsim.inputs import WindPowerPlantInputs
from pysimmods.generator.windsim.state import WindPowerPlantState
from pysimmods.util.air_density_at_height import (
    AIR_DENSITY_KG_PER_M3,
    barometric,
    ideal_gas,
)
from pysimmods.util.temperature_at_height import linear_gradient
from pysimmods.util.wind_at_height import hellmann, logarithmic_profile


class WindPowerPlant(
    Generator[WindPowerPlantConfig, WindPowerPlantState, WindPowerPlantInputs]
):
    """Simulation model of a windturbine plant.

    The code for the corrections and power output is heavily inspired
    by the windpowerlib::

        https://github.com/wind-python/windpowerlib

    """

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config: WindPowerPlantConfig = WindPowerPlantConfig(params)
        self.state: WindPowerPlantState = WindPowerPlantState(inits)
        self.inputs: WindPowerPlantInputs = WindPowerPlantInputs()

    def step(self, pretend: bool = False) -> WindPowerPlantState:
        """Perform a simulation step."""
        next_state = deepcopy(self.state)
        self._check_inputs(next_state)

        # First step: wind speed at hub height using hellmann or
        # logarithmic profile
        if self.config.wind_profile == "hellmann":
            next_state.wind_hub_v_m_per_s = hellmann(
                self.inputs.wind_v_m_per_s,
                self.config.wind_height_m,
                self.config.hub_height_m,
            )

        elif self.config.wind_profile == "logarithmic":
            next_state.wind_hub_v_m_per_s = logarithmic_profile(
                self.inputs.wind_v_m_per_s,
                self.config.wind_height_m,
                self.config.hub_height_m,
                self.config.roughness_length_m,
                self.config.obstacle_height_m,
            )

        # Second step: temperature at hub height using linear gradient
        # need T and P  at hub height
        if self.config.temperature_profile == "linear_gradient":
            next_state.t_air_hub_deg_kelvin = linear_gradient(
                self.inputs.t_air_deg_kelvin,
                self.config.temperature_height_m,
                self.config.hub_height_m,
            )
        else:
            next_state.t_air_hub_deg_kelvin = self.inputs.t_air_deg_kelvin

        # Third step: air density at hub height using temperature at
        # hub height. Two different profiles available: barometric or
        # ideal_gas

        if self.config.air_density_profile == "barometric":
            next_state.air_density_hub_kg_per_m3 = barometric(
                self.inputs.air_pressure_hpa,
                next_state.t_air_hub_deg_kelvin,
                self.config.pressure_height_m,
                self.config.hub_height_m,
            )
            wind_speed_corrected = self.wind_curve_correction(
                next_state.air_density_hub_kg_per_m3
            )

        elif self.config.air_density_profile == "ideal_gas":
            next_state.air_density_hub_kg_per_m3 = ideal_gas(
                self.inputs.air_pressure_hpa,
                next_state.t_air_hub_deg_kelvin,
                self.config.pressure_height_m,
                self.config.hub_height_m,
            )
            wind_speed_corrected = self.wind_curve_correction(
                next_state.air_density_hub_kg_per_m3
            )
        else:
            # No air density correction; use default value
            next_state.air_density_hub_kg_per_m3 = AIR_DENSITY_KG_PER_M3
            wind_speed_corrected = self.config.wind_speeds_m_per_s

        # Last step: power output with either power_curve or
        # power_coefficient method
        if self.config.method == "power_curve":
            next_state.p_kw = self.power_curve_output(
                next_state.wind_hub_v_m_per_s, wind_speed_corrected
            )
        elif self.config.method == "power_coefficient":
            next_state.p_kw = self.power_coefficient_output(
                next_state.air_density_hub_kg_per_m3,
                next_state.wind_hub_v_m_per_s,
            )

        if not pretend:
            self.state = next_state

        return next_state

    def wind_curve_correction(
        self, density: float, density0: float = AIR_DENSITY_KG_PER_M3
    ) -> np.ndarray:
        """Power curve at hub height

        Air density initial is set to the default value 1.225
        The correction is displayed in the new values of
        power_curve_wind_speeds

        Parameters
        ----------
        density: float
            Air density at hub height in [kg/m^3].
        density0: float
            Reference air density in [kg/m^3], defaults to 1.255.

        Returns
        -------
        np.ndarray
            Corrected wind speed values for given air density.
        """
        interpolation = np.interp(
            self.config.wind_speeds_m_per_s, [7.5, 12.5], [1 / 3, 2 / 3]
        )
        wind_curve_corrected = (
            np.array(density0 / density).reshape(-1, 1) ** interpolation
        ) * self.config.wind_speeds_m_per_s

        return wind_curve_corrected[0]

    def power_curve_output(
        self, wind_speed: float, wind_curve: np.ndarray
    ) -> float:
        """Calculate power output using power curve methond.

        Calculation of the power curve value for a given wind speed by
        interpolation of the `power_curve` array.

        Parameters
        ----------
        wind_speed: float
            The wind speed at hub height in [m/s].
        wind_curve: np.ndarray
            The wind curve at hub height in [m/s], used for interpolation
            with the power curve stored in the model config.

        Returns
        -------
        float
            Power output at the given wind_speed in [kW].

        """

        power_output = np.interp(
            wind_speed, wind_curve, self.config.power_curve_w, left=0, right=0
        )
        return float(power_output / 1000.0)  # to kW

    def power_coefficient_output(self, air_density: float, wind_speed: float):
        """Calculate power output using the power coffecient method.

        The method uses more configuration parameters of the model,
        including rotor_diameter_m and air density, which makes it
        more versatile than the power curve method.

        Parameters
        ----------
        air_density: float
            Air density
        wind_speed: float
            Wind speed at hub height in [m/s]

        Returns
        -------
        float
            Power output at the given wind_speed in [kW].

        """

        power_coefficient_inter = np.interp(
            wind_speed,
            self.config.wind_speeds_m_per_s,
            self.config.power_coefficient,
            left=0,
            right=0,
        )

        return (
            float(
                1
                / 8
                * air_density
                * self.config.rotor_diameter_m**2
                * np.pi
                * np.power(wind_speed, 3)
                * power_coefficient_inter
            )
            / 1000.0
        )  # to kW

    def _check_inputs(self, nstate):
        pass



"""This module contains the state model for the PV system."""

from typing import cast

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class PVSystemState(ModelState):
    """State variables of PV plant system model.

    Parameters
    ----------
    inits : dict
        A *dict* containing initialization parameters.

    Attributes
    ----------
    t_module_deg_celsius : float
        See :attr:`~.PVState.t_module_deg_celsius`
    p_kw : float
        See :attr:`~.ModelState.p_kw`
    q_kvar : float
        See :attr:`~.ModelState.q_kvar`
    p_possible_max_kw: float
        Power output of the PV module before inverter gets active.
    inverter_inductive: int
        1 if inverter mode was inductive, else 0
    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.t_module_deg_celsius: float = cast(
            "float", cast("ModelInitVals", inits["pv"])["t_module_deg_celsius"]
        )
        self.p_possible_max_kw: float = 0.0
        self.cos_phi: float = 0.0
        self.inverter_inductive: int = 0



"""This module contains multiple configuration examples for
wind_turbines with different nominal power.

"""

import csv
import logging
import os
from typing import cast

import numpy as np

from pysimmods.base.types import ModelInitVals, ModelParams

LOG = logging.getLogger(__name__)


def wind_preset(
    pn_max_kw: float,
    wind_profile: str = "hellmann",
    temperature_profile: str = "linear_gradient",
    air_density_profile: str = "barometric",
    power_method: str = "power_curve",
    turbine_type: str | None = None,
    randomize_params: bool = False,
    seed: int | None = None,
    **kwargs,
) -> tuple[ModelParams, ModelInitVals]:
    """Return the parameter configuration for a turbine model"""

    rng = np.random.default_rng(seed)
    data = load_static_data()

    if wind_profile not in ["hellmann", "logarithmic"]:
        wind_profile = "hellmann"

    if temperature_profile not in ["linear_gradient", "no_profile"]:
        temperature_profile = "linear_gradient"

    if air_density_profile not in ["barometric", "ideal_gas", "no_profile"]:
        air_density_profile = "barometric"

    if power_method not in ["power_curve", "power_coefficient"]:
        power_method = "power_curve"

    params: ModelParams = {}
    # Check if a matching model already exists
    if (
        turbine_type is not None
        and turbine_type in data
        and (
            power_method == "power_curve"
            and data[turbine_type]["has_power_curve"]
            or power_method == "power_coefficient"
            and data[turbine_type]["has_power_coefficient"]
        )
    ):
        params = cast("ModelParams", data[turbine_type])

    if not params:
        # No matching model found. Looking for the closest
        # model
        possible_pn = {}
        for ttype, values in data.items():
            possible_pn.setdefault(values["pn_max_kw"], [])
            if power_method == "power_curve" and values["has_power_curve"]:
                possible_pn[values["pn_max_kw"]].append(ttype)
            if (
                power_method == "power_coefficient"
                and values["has_power_coefficient"]
            ):
                possible_pn[values["pn_max_kw"]].append(ttype)

        if pn_max_kw in possible_pn:
            # A model for the given nominal power exists
            if len(possible_pn[pn_max_kw]) == 1:
                params = data[possible_pn[pn_max_kw][0]]
            else:
                if (
                    turbine_type is not None
                    and turbine_type in possible_pn[pn_max_kw]
                ):
                    params = data[turbine_type]
                elif randomize_params:
                    rnd = str(rng.choice(possible_pn[pn_max_kw]))
                    params = cast("ModelParams", data[rnd])
                else:
                    params = data[possible_pn[pn_max_kw][0]]
        else:
            # No model for the given nominal power exists
            # Look for the closest one and scale the power
            # values.
            p_lower = 0
            lower_diff = 0
            p_higher = 0
            higher_diff = 0
            for pn, ttypes in possible_pn.items():
                if not ttypes:
                    continue
                if pn < pn_max_kw:
                    p_lower = pn
                    lower_diff = pn_max_kw - pn
                if pn > pn_max_kw:
                    p_higher = pn
                    higher_diff = pn - pn_max_kw
                    break
            if (
                lower_diff < higher_diff
                and p_lower in possible_pn
                or lower_diff > higher_diff
                and p_higher not in possible_pn
            ):
                params = cast("ModelParams", data[possible_pn[p_lower][0]])

            else:
                params = data[possible_pn[p_higher][0]]
            params["pn_max_kw"] = pn_max_kw

    if power_method == "power_curve":
        params["power_curve_m"] = [
            v * 1000 * pn_max_kw
            for v in cast("list[float]", params["power_curve"])
        ]

    if randomize_params:
        h_min = cast("int", params["height_min_m"])
        h_max = cast("int", params["height_max_m"])
        params["hub_height_m"] = int(rng.integers(h_min, h_max))
    else:
        params["hub_height_m"] = params["height_min_m"]

    params["sign_convention"] = "active"
    params["method"] = power_method
    params["wind_profile"] = wind_profile
    params["temperature_profile"] = temperature_profile
    params["air_density_profile"] = air_density_profile
    return params, {}


def load_static_data():
    data_path = os.path.abspath(
        os.path.join(__file__, "..", "static_data.csv")
    )
    curves_path = os.path.abspath(
        os.path.join(__file__, "..", "static_curves.csv")
    )
    data = {}
    cols = []
    with open(data_path, "r") as csvfile:
        reader = csv.reader(csvfile)
        for idx, row in enumerate(reader):
            if idx == 0:
                cols = row
            else:
                key = row[0]
                data[key] = {}
                for k, v in zip(cols, row):
                    try:
                        v = int(v)
                    except ValueError:
                        try:
                            v = float(v)
                        except ValueError:
                            if v == "false":
                                v = False
                            elif v == "true":
                                v = True
                    data[key][k] = v
    with open(curves_path, "r") as csvfile:
        reader = csv.reader(csvfile)
        for idx, row in enumerate(reader):
            if idx == 0:
                cols = row
            else:
                key = row[0]
                ts_type = row[1]
                ts = []
                for v in row[2:]:
                    ts.append(float(v))
                data[key][ts_type] = ts
    return data



from enum import Enum
from typing import Generic

from typing_extensions import override

from pysimmods.base.generator import Generator
from pysimmods.base.model import C, I, S


class QGenerator(Generator[C, S, I], Generic[C, S, I]):
    """A generator subtype model with reactive power support."""

    @override
    def set_q_kvar(self, q_kvar: float | None) -> None:
        """Set target value for reactive power in kVAr.

        Can be set to None to fall back to the default schedule.
        """
        if q_kvar is None:
            self.inputs.q_set_kvar = None
        else:
            self.inputs.q_set_kvar = q_kvar * self.config.gsign

    @override
    def get_q_kvar(self) -> float:
        """Return current reactive power in kVAr of the model.

        Respects the current sign convention.
        """
        return self.state.q_kvar * self.config.gsign

    def get_qn_max_kvar(self) -> float:
        """Return the nominal maximum reactive power output in kVAr."""
        return self.config.s_max_kva



from copy import deepcopy
from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.inputs import ModelInputs
from pysimmods.base.qgenerator import QGenerator
from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals, ModelParams, QControl


class DummyQGeneratorConfig(ModelConfig):
    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.p_max_kw: float = cast("float", params.get("p_max_kw", 500))
        self.p_min_kw: float = cast("float", params.get("p_min_kw", 250))
        self.q_min_kvar: float = cast("float", params.get("q_min_kvar", -550))
        self.q_max_kvar: float = cast("float", params.get("q_max_kvar", 550))
        self.default_p_schedule = [375.0] * 24
        self.default_q_schedule = [125.0] * 24
        self.s_max_kva: float = cast(
            "float", params.get("s_max_kva", self.p_max_kw * 1.2)
        )
        self.q_control: QControl = cast(
            "QControl", params.get("q_control", QControl.PRIORITIZE_P)
        )
        self.cos_phi: float = cast("float", params.get("cos_phi", 0.9))


class DummyQGeneratorState(ModelState):
    pass


class DummyQGeneratorInputs(ModelInputs):
    def __init__(self):
        super().__init__()

        self.cos_phi: float | None = None


class DummyQGenerator(
    QGenerator[
        DummyQGeneratorConfig, DummyQGeneratorState, DummyQGeneratorInputs
    ]
):
    """The dummy Q generator model."""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = DummyQGeneratorConfig(params)
        self.state = DummyQGeneratorState(inits)
        self.inputs = DummyQGeneratorInputs()

    def step(self, pretend: bool = False) -> DummyQGeneratorState:
        next_state = deepcopy(self.state)

        # Check inputs
        next_state.p_kw = (
            0.0 if self.inputs.p_set_kw is None else self.inputs.p_set_kw
        )
        next_state.q_kvar = (
            0.0 if self.inputs.q_set_kvar is None else self.inputs.q_set_kvar
        )

        # Apply constraints
        next_state.p_kw = min(
            self.config.p_max_kw, max(self.config.p_min_kw, next_state.p_kw)
        )
        next_state.q_kvar = min(
            self.config.s_max_kva,
            max(-self.config.s_max_kva, next_state.q_kvar),
        )

        if not pretend:
            self.state = next_state
        return next_state



"""This module contains the base config used by each model."""

from pysimmods.base.types import ModelParams


class ModelConfig:
    """The base config for all models.

    Parameters
    ----------
    params: dict
        A *dict* containing configuration parameters. See *Attributes*
        for detailed information

    Attributes
    ----------
    sign_convention: str
        Should be defined in *params* and can have the values
        *'passive'* or *'active'*. Passive sign convention aka load
        reference arrow system is normally used and leads to positive
        consumption and negative generation power flows.
    psc: bool
        Will be automatically set to *True* if passive sign convention
        is used.
    asc: bool
        Will be automatically set to *True* if active sign convention is
        used.
    gsign: int
        The generator sign depending on the sign convention. Will be
        set to *-1* if passive sign convention is used and to *1*
        otherwise.
    lsign: int
        The load sign depending on the sign convention. Will be set to
        *1* is passive sign convention is used and to *-1* otherwise.
    use_decimal_percent: bool
        If *True* decimal percentage [0, 1.0] is used, otherwise normal
        percentage [0, 100] ist used.
    default_p_schedule: List[float]
        A *list* containing a default schedule for active power for
        each hour of the day (i.e., len(default_schedule) == 24). This
        is used if no other *p* input is provided.
    default_q_schedule: List[float]
        A *list* containing a default schedule for reactive power for
        each hour of the day (i.e., len(default_schedule) == 24). This
        is used if no other *q* input is provided.
    p_min_kw: float
        The minimum nominal active power output of the model.
    p_max_kw: float
        The maximum nominal active power output of the model.
    q_min_kvar: float
        The minimum nominal reactive power output of the model.
    q_max_kvar: float
        The maximum nominal reactive power output of the model.
    s_min_kva: float
        The minimal nominal apparent power output of the model.
    s_max_kva: float
        The maximal nominal apparent power output of the model.

    """

    def __init__(self, params: ModelParams) -> None:
        self.sign_convention = params.get("sign_convention", "passive")
        if self.sign_convention not in ["active", "passive"]:
            msg = (
                f"Invalid sign convention {self.sign_convention}. "
                "Allowed values are 'active' and 'passive'"
            )
            raise ValueError(msg)

        self.psc = self.sign_convention == "passive"
        self.asc = self.sign_convention == "active"

        if self.sign_convention == "active":
            self.gsign = 1
            self.lsign = -1
        else:
            self.gsign = -1
            self.lsign = 1

        self.use_decimal_percent = params.get("use_decimal_percent", False)
        self.default_p_schedule = params.get("default_p_schedule", [0.0] * 24)
        self.default_q_schedule = params.get("default_q_schedule", [0.0] * 24)

        self.p_min_kw: float = 0.0
        self.p_max_kw: float = 0.0
        self.q_min_kvar: float = 0.0
        self.q_max_kvar: float = 0.0
        self.s_min_kva: float = 0.0
        self.s_max_kva: float = 0.0



"""This module contains the config model for the chp system."""

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams
from pysimmods.generator.chplpgsim.config import CHPLPGConfig
from pysimmods.other.heatdemandsim.config import HeatDemandConfig


class CHPLPGSystemConfig(ModelConfig):
    """captures the configs for the chp system's components"""

    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.house = HeatDemandConfig(params.get("household", {}))
        self.chp = CHPLPGConfig(params["chp"])
        self.p_max_kw = self.chp.p_max_kw
        self.p_min_kw = self.chp.p_min_kw
        self.q_min_kvar = self.chp.q_min_kvar
        self.q_min_kvar = self.chp.q_max_kvar
        self.e_th_demand_sign = 1.0
        if params.get("flip_e_th_demand_sign", False):
            self.e_th_demand_sign = -1.0
        self.default_schedule = self.chp.default_schedule



"""This module contains the state model of the chp."""

from typing import cast

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class CHPLPGState(ModelState):
    """A CHP LGP state information class

    This class captures the state of the chp model. States normally
    change during the step-method of this model.

    Parameters
    ----------
    inits : dict
        A *dict* containing initialization parameters for the model.

    Attributes
    ----------
    active_s : int
        Time since unit was switched on in [s].
    inactive_s : int
        Time since unit was switch off in [s].
    is_active : bool
        True if the unit is operating.
    storage_t_c : float
        Stores the current temperature of the heat storage in [°C].
    lubricant_l : float, optional
        Current amount of lubricant in [l]. Defaults to 10.

    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.active_s: int = cast("int", inits["active_s"])
        self.inactive_s: int = cast("int", inits["inactive_s"])
        self.is_active: int = cast("int", inits["is_active"])
        self.storage_t_c: float = cast("float", inits["storage_t_c"])
        self.lubricant_l: float = cast("float", inits.get("lubricant_l", 10))
        self.p_th_kw: float = 0.0



"""This module contains a model of a wind turbine and an inverter"""

from copy import deepcopy
from typing import cast

from pysimmods.base.qgenerator import QGenerator
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.generator.windsim.wind import WindPowerPlant
from pysimmods.generator.windsystemsim.config import WindSystemConfig
from pysimmods.generator.windsystemsim.inputs import WindSystemInputs
from pysimmods.generator.windsystemsim.state import WindSystemState
from pysimmods.other.invertersim.inverter import Inverter


class WindPowerPlantSystem(
    QGenerator[WindSystemConfig, WindSystemState, WindSystemInputs]
):
    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = WindSystemConfig(params)
        self.inputs = WindSystemInputs()
        self.state = WindSystemState(inits)

        self.wind = WindPowerPlant(
            cast("ModelParams", params["wind"]),
            cast("ModelInitVals", inits.get("wind", {})),
        )
        self.inverter = Inverter(
            cast("ModelParams", params["inverter"]),
            cast("ModelInitVals", inits.get("inverter", {})),
        )

    def step(self, pretend: bool = False) -> WindSystemState:
        """Perform simulation step."""
        next_state = deepcopy(self.state)

        # Step the wind turbine
        # self.wind.inputs.step_size = self.inputs.step_size
        # self.wind.inputs.now_dt = self.inputs.now_dt
        self.wind.inputs.wind_v_m_per_s = self.inputs.wind_v_m_per_s
        self.wind.inputs.t_air_deg_celsius = self.inputs.t_air_deg_celsius
        self.wind.inputs.air_pressure_hpa = self.inputs.air_pressure_hpa
        wind_state = self.wind.step(pretend)

        # Step the inverter
        self.inverter.inputs.p_in_kw = wind_state.p_kw
        self.inverter.inputs.p_set_kw = self.inputs.p_set_kw
        self.inverter.inputs.q_set_kvar = self.inputs.q_set_kvar
        self.inverter.inputs.cos_phi_set = self.inputs.cos_phi_set
        self.inverter.inputs.inductive = self.inputs.inverter_inductive
        inv_state = self.inverter.step(pretend)

        # Update state
        next_state.p_kw = inv_state.p_kw
        next_state.p_possible_max_kw = wind_state.p_kw
        next_state.q_kvar = inv_state.q_kvar
        next_state.cos_phi = inv_state.cos_phi
        next_state.inverter_inductive = inv_state._inductive
        next_state.wind_hub_v_m_per_s = wind_state.wind_hub_v_m_per_s
        next_state.t_air_hub_deg_kelvin = wind_state.t_air_hub_deg_kelvin
        next_state.air_density_hub_kg_per_m3 = (
            wind_state.air_density_hub_kg_per_m3
        )

        if not pretend:
            self.state = next_state

        return next_state

    def get_state(self):
        state_dict = {
            "wind": self.wind.get_state(),
            "inverter": self.inverter.get_state(),
        }
        return state_dict

    def set_state(self, state):
        self.wind.set_state(state["wind"])
        self.inverter.set_state(state["inverter"])

        self.state.p_kw = self.inverter.state.p_kw
        self.state.p_possible_max_kw = self.wind.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state._inductive
        self.state.wind_hub_v_m_per_s = self.wind.state.wind_hub_v_m_per_s
        self.state.t_air_hub_deg_kelvin = self.wind.state.t_air_hub_deg_kelvin
        self.state.air_density_hub_kg_per_m3 = (
            self.wind.state.air_density_hub_kg_per_m3
        )

    # def set_q_kvar(self, q_kvar: float):
    #     self.inputs.q_set_kvar = q_kvar



"""This module contains the input model for the chp system."""

from pysimmods.base.inputs import ModelInputs


class CHPLPGSystemInputs(ModelInputs):
    """captures the inputs of the chp system"""

    def __init__(self):
        super().__init__()

        self.day_avg_t_air_deg_celsius: float = 0.0
        """Average temperature of the current day in [°C]."""

        self.e_th_demand_set_kwh: float | None = None



import numpy as np


def logarithmic_profile(
    wind_v_m_per_s: float,
    wind_height_m: float,
    hub_height_m: float,
    roughness_length_m: float = 0.15,
    obstacle_height_m: float = 0.0,
) -> float:
    """Calculate wind speed at altitude using logarithm.

    Parameters
    ----------
    wind_v_m_per_s: float
        The measured wind speed in [m/s].
    wind_height_m: float
        The height on which the wind speed was measured in [m].
    hub_height_m: float
        The height for which the wind speed should be calculated in [m].
    roughness_length_m: float
        Length of roughness in [m].
    obstacle_height_m: float
        Height of obstacles that affect wind speed in [m].

    Parameters
    ----------
    float
        Wind speed at altitude `hub_height_m`.

    """
    return (
        wind_v_m_per_s
        * np.log((hub_height_m - 0.7 * obstacle_height_m) / roughness_length_m)
        / np.log(
            (wind_height_m - 0.7 * obstacle_height_m) / roughness_length_m
        )
    )


def hellmann(
    wind_v_m_per_s: float,
    wind_height_m: float,
    hub_height_m: float,
    hellman_exponent: float = 1 / 7,
) -> float:
    """Calculate wind speed at altitude using hellmann formula

    Parameters
    ----------
    wind_v_m_per_s: float
        The measured wind speed in [m/s].
    wind_height_m: float
        The height on which the wind speed was measured in [m].
    hub_height_m: float
        The height for which the wind speed should be calculated in [m].
    hellmann_exponent: float
        The hellmann exponent for this calculation, defaults to 1/7.

    Returns
    -------
    float
        Wind speed at altitude `hub_height_m`.
    """
    return wind_v_m_per_s * (hub_height_m / wind_height_m) ** hellman_exponent



"""This module contains the config model for the Wind plant."""

from typing import cast

import numpy as np

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class WindPowerPlantConfig(ModelConfig):
    """Config parameters of Wind plant model.

    Parameters
    ----------

    params : dict
        Contains the configuration of the turbine. See attribute
        section for more information about the parameters, attributes
        marked with *(Input)* can or must be provided.

    Attributes
    ----------

    pn_max_kw : float
        Nominal power of the wind turbine in [kW].
    turbine_type: str
        Model type of this turbine. Has no practical impact.
    hub_height_m: float
        Height of the wind turbine in [m].
    rotor_diameter_m: float
        Diameter of the rotor of the turbine in [m].
    roughness_length_m: float
        Length of roughness in [m].
    obstacle_height_m: float
        Height of obstacles that affect wind speed in [m].
    wind_height_m: float
        Height in which wind speed was measured in [m].
    temperature_height_m: float
        Height in which temperature was measured in [m].
    pressure_height_m: float
        Height in which air pressure was measured in [m].
    method: str
        The method used to calculate wind power. Can be either
        `power_curve` or `power_coefficient`.
    temperature_profile: str
        Method to extrapolate the temperature at hub height. Can be
        either `linear_gradient` or `no_profile` to directly use the
        provided temperature values without extrapolation.
    wind_profile: str
        Method to calculate wind speed at hub height. Can be either
        `hellmann` or `logarithmic`.
    air_density_profile: str
        Method to calculate air density at hub height. Can be either
        `barometric`, `ideal_gas`, or `no_profile`.
    power_curve_w: np.ndarray
        Contains the power values used for power curve calculation
        in [w].
    power_coefficient: np.ndarray
        Contains the values used for power coefficient calculation.
    wind_speeds_m_per_s: np.ndarray
        Contains the wind speed levels for either `power_curve` or
        `power_coefficient`, depending on which `method` is used.

    """

    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.p_max_kw: float = cast("float", params.get("pn_max_kw", 2000.0))
        self.p_min_kw: float = 0
        self.turbine_type: str = cast(
            "str", params.get("turbine_type", "E-82/2000")
        )
        self.hub_height_m: float = cast(
            "float", params.get("hub_height_m", 78)
        )
        self.rotor_diameter_m: float = cast(
            "float", params.get("rotor_diameter_m", 82)
        )
        self.roughness_length_m: float = cast(
            "float", params.get("roughness_length_m", 0.15)
        )
        self.obstacle_height_m: float = cast(
            "float", params.get("obstacle_height_m", 0)
        )

        self.wind_height_m: float = cast(
            "float", params.get("wind_height_m", 10)
        )
        self.temperature_height_m: float = cast(
            "float", params.get("temperature_height_m", 10)
        )
        self.pressure_height_m: float = cast(
            "float", params.get("pressure_height_m", 10)
        )

        self.method: str = cast("str", params.get("method", "power_curve"))

        self.temperature_profile: str = cast(
            "str", params.get("temperature_profile", "linear_gradient")
        )
        if self.temperature_profile != "linear_gradient":
            self.temperature_profile = "no_profile"

        self.wind_profile: str = cast(
            "str", params.get("wind_profile", "hellmann")
        )

        self.air_density_profile: str = cast(
            "str", params.get("air_density_profile", "barometric")
        )
        self.power_curve_w: np.ndarray = np.array(
            cast(
                "list[float]",
                params.get("power_curve_m", [0, 26, 180, 1500, 3000, 3000]),
            )
        )

        self.power_coefficient: np.ndarray = np.array(
            cast(
                "list[float]",
                params.get(
                    "power_coefficient",
                    [0.301, 0.36, 0.39, 0.409, 0.421, 0.429],
                ),
            )
        )

        self.wind_speeds_m_per_s = np.array(params.get("wind_speeds", []))
        if self.wind_speeds_m_per_s.size == 0:
            if self.method == "power_curve":
                num_vals = len(self.power_curve_w)
            else:
                num_vals = len(self.power_coefficient)

            self.wind_speeds_m_per_s = np.linspace(0.0, 25.0, num_vals)



from copy import deepcopy
from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.consumer import Consumer
from pysimmods.base.inputs import ModelInputs
from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals, ModelParams


class DummyConsumerConfig(ModelConfig):
    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.p_max_kw: float = cast("float", params.get("p_max_kw", 500))
        self.p_min_kw: float = cast("float", params.get("p_min_kw", 250))
        self.q_max_kvar: float = cast("float", params.get("q_max_kvar", 100))
        self.q_min_kvar: float = cast("float", params.get("q_min_kvar", 25))

        self.default_p_schedule = [10.0] * 24
        self.default_q_schedule = [-4.0] * 24


class DummyConsumerState(ModelState):
    pass


class DummyConsumerInputs(ModelInputs):
    pass


class DummyConsumer(
    Consumer[DummyConsumerConfig, DummyConsumerState, DummyConsumerInputs]
):
    """The dummy consumer model."""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = DummyConsumerConfig(params)
        self.state = DummyConsumerState(inits)
        self.inputs = DummyConsumerInputs()

    def step(self, pretend=False) -> DummyConsumerState:
        next_state = deepcopy(self.state)

        next_state.p_kw = (
            0.0 if self.inputs.p_set_kw is None else self.inputs.p_set_kw
        )
        if next_state.p_kw > 0:
            next_state.p_kw = min(
                self.config.p_max_kw,
                max(self.config.p_min_kw, next_state.p_kw),
            )
        if not pretend:
            self.state = next_state
        return next_state



from copy import deepcopy
from typing import cast

from pysimmods.base.buffer import Buffer
from pysimmods.base.config import ModelConfig
from pysimmods.base.inputs import ModelInputs
from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals, ModelParams


class DummyBufferConfig(ModelConfig):
    def __init__(self, params: ModelParams):
        params.setdefault("default_schedule", [10.0] * 24)
        super().__init__(params)

        self.p_max_kw: float = cast("float", params.get("p_max_kw", 500))
        self.p_min_kw: float = cast("float", params.get("p_min_kw", 250))
        self.q_max_kvar: float = cast("float", params.get("q_max_kvar", 100))
        self.q_min_kvar: float = cast("float", params.get("q_min_kvar", 25))
        self.cap_e_khw: float = cast("float", params.get("cap_e_kwh", 5000))

        self.p_charge_max_kw: float = abs(
            cast("float", params.get("p_charge_max_kw", self.p_max_kw))
        )
        self.p_charge_min_kw: float = abs(
            cast("float", params.get("p_charge_min_kw", self.p_min_kw))
        )

        self.p_discharge_max_kw: float = abs(
            cast("float", params.get("p_discharge_max_kw", -self.p_max_kw))
        )

        self.p_discharge_min_kw: float = abs(
            cast("float", params.get("p_discharge_min_kw", -self.p_min_kw))
        )


class DummyBufferState(ModelState):
    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)
        self.soc_kwh: float = cast("float", inits["soc_kwh"])


class DummyBufferInputs(ModelInputs):
    pass


class DummyBuffer(
    Buffer[DummyBufferConfig, DummyBufferState, DummyBufferInputs]
):
    """The dummy buffer model."""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = DummyBufferConfig(params)
        self.state = DummyBufferState(inits)
        self.inputs = DummyBufferInputs()

    def step(self, pretend=False) -> DummyBufferState:
        next_state = deepcopy(self.state)

        # Check inputs
        step_size = (
            1 if self.inputs.step_size is None else self.inputs.step_size
        )
        next_state.p_kw = (
            0.0 if self.inputs.p_set_kw is None else self.inputs.p_set_kw
        )
        # Apply constraints
        if next_state.p_kw < 0:
            next_state.p_kw = min(
                -self.config.p_discharge_min_kw,
                max(-self.config.p_discharge_max_kw, next_state.p_kw),
            )
        elif next_state.p_kw > 0:
            next_state.p_kw = min(
                self.config.p_charge_max_kw,
                max(self.config.p_charge_min_kw, next_state.p_kw),
            )

        e_khw = next_state.p_kw * step_size / 3600
        if next_state.soc_kwh + e_khw > self.config.cap_e_khw:
            next_state.p_kw = 0
        elif next_state.soc_kwh + e_khw < 0:
            next_state.p_kw = 0
        else:
            next_state.soc_kwh += e_khw

        if not pretend:
            self.state = next_state

        return next_state



"""This module contains the flexibility model."""

import logging
from datetime import datetime, timedelta, timezone
from typing import Optional

import numpy as np
import pandas as pd

from pysimmods.base.model import I
from pysimmods.other.flexibility.flexibilities import Flexibilities
from pysimmods.other.flexibility.forecast_model import ForecastModel
from pysimmods.other.flexibility.schedule import Schedule
from pysimmods.util.date_util import GER

LOG = logging.getLogger(__name__)


class FlexibilityModel(ForecastModel):
    """The flexibility model for all pysimmods."""

    def __init__(
        self,
        model,
        unit="kw",
        prioritize_setpoint: bool = False,
        step_size: Optional[int] = None,
        now_dt: Optional[datetime] = None,
        forecast_horizon_hours: float = 1.0,
        seed: int | None = None,
        store_min_and_max: bool = False,
    ):
        super().__init__(
            model,
            unit,
            prioritize_setpoint,
            step_size,
            now_dt,
            forecast_horizon_hours,
        )

        self._store_min_and_max = store_min_and_max
        self._rng = np.random.default_rng(seed)

    def generate_schedules(
        self, start: str | datetime, flexibility_horizon_hours, num_schedules
    ) -> Flexibilities:
        """Perform sampling and generate a set of schedules for the
        specified time interval.

        Args:
            start (str): Is the start of the planning horizon for which
                the sampling is done. It has to be provided as ISO 8601
                timezone string such as '2020-06-22 12:00:00+0000'

            forecast_horizon_hours (int): The planning horizon is divided into.

        """
        state_backup = self._model.get_state()
        step_size = self._step_size

        now_dt = self._now_dt
        if isinstance(start, str):
            start_dt = datetime.strptime(start, GER).astimezone(timezone.utc)
        else:
            start_dt = start
        end_dt = (
            start_dt
            + timedelta(hours=flexibility_horizon_hours)
            - timedelta(seconds=step_size)
        )
        periods = int(flexibility_horizon_hours * 3_600 / step_size)

        # Fast forward to the planning interval
        while start_dt > now_dt:
            try:
                p_set = self.schedule.get(now_dt, self._psetname)
                if p_set is None or np.isnan(p_set):
                    # Fallback to schedule that was implicitly
                    # created
                    p_set = self.schedule.get(now_dt, self._pname)
                q_set = self.schedule.get(now_dt, self._qsetname)
                if q_set is None or np.isnan(q_set):
                    q_set = self.schedule.get(now_dt, self._qname)
                self._calculate_step(now_dt, p_set, q_set)
                now_dt += timedelta(seconds=step_size)
            except KeyError as err:
                # raise err
                LOG.info("Could not create flexibilities: %s", err)
                return {}

        ff_index = pd.date_range(
            self._now_dt,
            start_dt - timedelta(seconds=step_size),
            freq=f"{step_size}s",
        )
        index = pd.date_range(start_dt, end_dt, periods=periods)

        self.flexibilities = Flexibilities()

        for schedule_id in range(num_schedules):
            self.flexibilities.add_schedule(
                schedule_id, self.sample(index, ff_index)
            )

        self._model.set_state(state_backup)
        return self.flexibilities

    def sample(self, index, ff_index):
        if not ff_index.empty:
            ff_df = pd.DataFrame(
                columns=[
                    self._psetname,
                    self._qsetname,
                    self._pname,
                    self._qname,
                ],
                index=ff_index,
                dtype="float",
            )
            for i, row in ff_df.iterrows():
                p_set = self.schedule.get(i, self._psetname)
                if p_set is None or np.isnan(p_set):
                    p_set = self.schedule.get(i, self._pname)
                q_set = self.schedule.get(I, self._qsetname)
                if q_set is None or np.isnan(q_set):
                    q_set = self.schedule.get(i, self._qname)
                ff_df.at[i, self._psetname] = p_set
                ff_df.at[i, self._qsetname] = q_set
        else:
            ff_df = pd.DataFrame()

        dataframe = pd.DataFrame(
            columns=[self._psetname, self._qsetname, self._pname, self._qname],
            index=index,
            dtype="float",
        )
        dataframe[self._psetname] = self._rng.uniform(
            low=self.get_pn_min_kw(),
            high=self.get_pn_max_kw(),
            size=len(index),
        )
        dataframe[self._qsetname] = self._rng.uniform(
            low=self.get_qn_min_kvar(),
            high=self.get_qn_max_kvar(),
            size=len(index),
        )

        state_backup = self._model.get_state()
        for index, row in dataframe.iterrows():
            try:
                self._calculate_step(
                    index, row[self._psetname], row[self._qsetname]
                )
                dataframe.loc[index, self._pname] = (
                    self._model.get_p_kw() * self._unit_factor
                )
                dataframe.loc[index, self._qname] = (
                    self._model.get_q_kvar() * self._unit_factor
                )

            except KeyError:
                # Forecast is missing
                dataframe.loc[index, self._pname] = np.nan
                dataframe.loc[index, self._qname] = np.nan
                dataframe.loc[index, self._psetname] = np.nan
                dataframe.loc[index, self._qsetname] = np.nan
        self._model.set_state(state_backup)
        if not ff_df.empty:
            dataframe = pd.concat([ff_df, dataframe])
        return Schedule().from_dataframe(dataframe)



"""This module contains the state model of the Wind Turbine System."""

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class WindSystemState(ModelState):
    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.p_possible_max_kw: float = 0
        self.wind_hub_v_m_per_s: float = 0
        self.t_air_hub_deg_kelvin: float = 0
        self.air_density_hub_kg_per_m3: float = 0
        self.cos_phi: float = 0
        self.inverter_inductive: bool = False



from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class HeatStorageConfig(ModelConfig):
    """Captures the configuration parameters of the heat storage."""

    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.cap_l: float = cast("float", params["storage_cap_l"])
        self.consumption_kwh_per_day: float = cast(
            "float", params["storage_consumption_kwh_per_day"]
        )
        self.t_min_c: float = cast("float", params["storage_t_min_c"])
        self.t_max_c: float = cast("float", params["storage_t_max_c"])
        self.env_t_c: float = cast(
            "float", params.get("storage_env_t_c", 19.0)
        )



"""This module contains multiple configuration examples for the
biogas plant.

"""

from pathlib import Path
from typing import cast

import numpy as np
import pandas as pd

from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.generator.chpcngsim.presets import (
    chpcng_preset,
    chpcng_preset_by_name,
)

BASE_DATA = (Path(__file__) / ".." / "base_data.csv").resolve()


def biogas_preset(
    pn_max_kw: float,
    randomize_params: bool = False,
    randomize_inits: bool = False,
    seed: int | None = None,
    **kwargs,
) -> tuple[ModelParams, ModelInitVals]:
    """Return the parameter configuration for a biogas model from the
    pysimmods package.

    """
    rng = np.random.default_rng(seed)

    all_params, data = _read_all_params()
    if randomize_params:
        params = {}

        params["gas_fill_min_percent"] = rng.uniform(
            data["gas_fill_min_percent"].min(),
            data["gas_fill_min_percent"].max(),
        )
        params["gas_fill_max_percent"] = rng.uniform(
            data["gas_fill_max_percent"].min(),
            data["gas_fill_max_percent"].max(),
        )
        params["ch4_concentration_percent"] = rng.uniform(
            data["ch4_concentration_percent"].min(),
            data["ch4_concentration_percent"].max(),
        )
        params["num_chps"] = int(rng.integers(1, 6))
        params["has_heat_storage"] = params["num_chps"] == 1

    else:
        closest_name = ""
        closest_distance = 2**32 - 1
        for n, params in all_params.items():
            pn_kw = int(n.split("_")[1].split("k")[0])
            dist = abs(pn_max_kw - pn_kw)
            if dist < closest_distance:
                closest_distance = dist
                closest_name = n

        params = all_params[closest_name]

    if randomize_inits:
        inits: ModelInitVals = {
            "gas_fill_percent": rng.uniform(
                cast("float", params["gas_fill_min_percent"]),
                cast("float", params["gas_fill_max_percent"]),
            )
        }
    else:
        inits: ModelInitVals = {"gas_fill_percent": 50.0}

    if randomize_params:
        chp_powers = _generate_split(rng, pn_max_kw, params["num_chps"])
        max_gas_con = 0.0
        for i, p in enumerate(chp_powers):
            cp, ci = chpcng_preset(
                p, randomize_params, randomize_inits, rng.integers(2**32 - 1)
            )
            cname = f"chp{i}"
            params[cname] = cp
            inits[cname] = ci
            max_gas_con += p / (
                cast("list[float]", cp["eta_stages_percent"])[-1]
                * 0.01
                * 9.94
                * cast("float", params["ch4_concentration_percent"])
                * 0.01
            )
        gas_cap_to_prod_ratio = rng.uniform(
            data["gas_cap_to_prod_ratio"].min(),
            data["gas_cap_to_prod_ratio"].max(),
        )
        gas_prod_to_con_ratio = rng.uniform(0.5, 0.8)
        params["gas_m3_per_day"] = int(
            max_gas_con * gas_prod_to_con_ratio * 24
        )
        params["cap_gas_m3"] = int(
            params["gas_m3_per_day"] * gas_cap_to_prod_ratio
        )
    else:
        max_gas_con = 0.0
        for i in range(3):
            cname = f"chp{i}"
            if cname not in params:
                break
            cp, ci = chpcng_preset_by_name(
                cast("str", params[cname]),
                randomize_inits,
                rng.integers(2**32 - 1),
            )
            params[cname] = cp
            inits[cname] = ci
            max_gas_con += cast("list[float]", cp["pn_stages_kw"])[-1] / (
                cast("list[float]", cp["eta_stages_percent"])[-1]
                * 0.01
                * 9.94
                * cast("float", params["ch4_concentration_percent"])
                * 0.01
            )

    params["sign_convention"] = "active"

    return params, inits


def _read_all_params() -> tuple[dict[str, ModelParams], pd.DataFrame]:
    data = pd.read_csv(BASE_DATA)
    all_params: dict[str, ModelParams] = {}

    for i, row in data.iterrows():
        p_kw = float(cast("str", row["name"]).split("_")[1].split("k")[0])
        data.at[i, "pn_kw"] = p_kw
        params: ModelParams = {
            "gas_m3_per_day": cast("float", row["gas_m3_per_day"]),
            "cap_gas_m3": cast("float", row["cap_gas_m3"]),
            "gas_fill_min_percent": cast("float", row["gas_fill_min_percent"]),
            "gas_fill_max_percent": cast("float", row["gas_fill_max_percent"]),
            "ch4_concentration_percent": cast(
                "float", row["ch4_concentration_percent"]
            ),
        }
        for j in range(3):
            name = row[f"chp{j}"]
            if isinstance(name, str):
                params[f"chp{j}"] = name
                params["num_chps"] = j + 1
            else:
                break
        all_params[cast("str", row["name"])] = params
    data["gas_cap_to_prod_ratio"] = data["cap_gas_m3"] / data["gas_m3_per_day"]
    return all_params, data


def _generate_split(rng: np.random.Generator, total, parts):
    if parts <= 1:
        return [total]

    raw = sorted([float(rng.uniform()) for _ in range(parts)], reverse=True)
    total_weight = sum(raw)
    weights = [w / total_weight for w in raw]

    values = [round(w * total) for w in weights]
    diff = total - sum(values)
    values[-1] += diff
    values.sort(reverse=True)
    return values


# def params_1g_40kw():
#     """Params for a Biogas plant with 40 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 550,
#         "cap_gas_m3": 250,
#         "gas_fill_min_percent": 2.5,
#         "gas_fill_max_percent": 97.5,
#         "ch4_concentration_percent": 51.9,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [20, 40],
#             "eta_stages_percent": [28.0, 30.0],
#             "eta_th_stages_percent": [31.4, 32.0],
#             "restarts_per_day": 0,
#             "active_min_s": 0,
#             "active_max_s_per_day": 18 * 3_600,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_40kw():
#     """Init vals for a Biogas plant with 40 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 20,
#         },
#     }


# def params_1g_80kw():
#     """Params for a Biogas plant with 80 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 750,
#         "cap_gas_m3": 375,
#         "gas_fill_min_percent": 2.5,
#         "gas_fill_max_percent": 97.5,
#         "ch4_concentration_percent": 52.9,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [40, 80],
#             "eta_stages_percent": [37.0, 36.7],
#             "eta_th_stages_percent": [32.4, 32.0],
#             "restarts_per_day": 5,
#             "active_min_s": 0,
#             "active_max_s_per_day": 18 * 3_600,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_80kw():
#     """Init vals for a Biogas plant with 80 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 40,
#         },
#     }


# def params_1g_320kw():
#     """Params for a Biogas plant with 320 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 2100,
#         "cap_gas_m3": 2000,
#         "gas_fill_min_percent": 5.0,
#         "gas_fill_max_percent": 95.0,
#         "ch4_concentration_percent": 55.3,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [160, 240, 320],
#             "eta_stages_percent": [36.6, 37.0, 37.5],
#             "eta_th_stages_percent": [31.0, 31.5, 31.9],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_320kw():
#     """Init vals for a Biogas plant with 320 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 160,
#         },
#     }


# def params_1g_550kw():
#     """Params for a Biogas plant with 550 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 5450,
#         "cap_gas_m3": 1578,
#         "gas_fill_min_percent": 2.0,
#         "gas_fill_max_percent": 98.0,
#         "ch4_concentration_percent": 55.4,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [275, 330, 440, 550],
#             "eta_stages_percent": [31.6, 31.8, 32.0, 32.2],
#             "eta_th_stages_percent": [41.0, 42.1, 43.2, 44.2],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_550kw():
#     """Init vals for a Biogas plant with 550 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 275,
#         },
#     }


# def params_1g_373kw():
#     """Params for a Biogas plant with 375 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 2550,
#         "cap_gas_m3": 2500,
#         "gas_fill_min_percent": 5.0,
#         "gas_fill_max_percent": 95.0,
#         "ch4_concentration_percent": 58.5,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [121, 252, 373],
#             "eta_stages_percent": [33.7, 36.2, 37.9],
#             "eta_th_stages_percent": [45.0, 46.0, 47.8],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_373kw():
#     """Init vals for a Biogas plant with 373 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 750,
#         },
#     }


# def params_1g_1250kw():
#     """Params for a Biogas plant with 1250 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_m3_per_day": 9_750,
#         "cap_gas_m3": 9_250,
#         "gas_fill_min_percent": 2.5,
#         "gas_fill_max_percent": 97.5,
#         "ch4_concentration_percent": 50.0,
#         "num_chps": 1,
#         "chp0": {
#             "pn_stages_kw": [750, 1250],
#             "eta_stages_percent": [37.0, 41.25],
#             "eta_th_stages_percent": [48.0, 44.0],
#             "restarts_per_day": 4,
#             "active_min_s": 0,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_1g_1250kw():
#     """Init vals for a Biogas plant with 1250 kw nominal power and
#     one CHP CNG.
#     """
#     return {
#         "gas_fill_percent": 50.0,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 750,
#         },
#     }


# def params_2g_55kw():
#     """Params for a Biogas plant with 55 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 255,
#         "cap_gas_m3": 390,
#         "gas_fill_min_percent": 5.0,
#         "gas_fill_max_percent": 95.0,
#         "ch4_concentration_percent": 50.2,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [9.0, 18.0],
#             "eta_stages_percent": [29.3, 28.7],
#             "eta_th_stages_percent": [35.4, 36.1],
#             "restarts_per_day": 0,
#             "active_min_s": 1 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [18.5, 37.0],
#             "eta_stages_percent": [31.4, 32.5],
#             "eta_th_stages_percent": [36.1, 36.8],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_55kw():
#     """Init vals for a Biogas plant with 55 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 27.5,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 18.3,
#         },
#     }


# def params_2g_110kw():
#     """Params for a Biogas plant with 110 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 950,
#         "cap_gas_m3": 1000,
#         "gas_fill_min_percent": 2.5,
#         "gas_fill_max_percent": 97.5,
#         "ch4_concentration_percent": 56.8,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [27.5, 55.0],
#             "eta_stages_percent": [29.7, 31.9],
#             "eta_th_stages_percent": [36.4, 37.0],
#             "restarts_per_day": 6,
#             "active_min_s": 1 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [18.3, 36.6, 55.0],
#             "eta_stages_percent": [30.8, 32.2, 33.9],
#             "eta_th_stages_percent": [36.1, 36.8, 38.0],
#             "restarts_per_day": 6,
#             "active_min_s": 1 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_110kw():
#     """Init vals for a Biogas plant with 110 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 27.5,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 18.3,
#         },
#     }


# def params_2g_220kw():
#     """Params for a Biogas plant with 220 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 2000,
#         "cap_gas_m3": 1150,
#         "gas_fill_min_percent": 2.5,
#         "gas_fill_max_percent": 97.5,
#         "ch4_concentration_percent": 55.7,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [55.0, 110.0],
#             "eta_stages_percent": [35.4, 34.7],
#             "eta_th_stages_percent": [31.4, 31.0],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [55.0, 110.0],
#             "eta_stages_percent": [35.5, 36.1],
#             "eta_th_stages_percent": [38.0, 42.4],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_220kw():
#     """Init vals for a Biogas plant with 220 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 55.0,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 55.0,
#         },
#     }


# def params_2g_300kw():
#     """Params for a Biogas plant with 300 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 3030,
#         "cap_gas_m3": 1530,
#         "gas_fill_min_percent": 2.0,
#         "gas_fill_max_percent": 98.0,
#         "ch4_concentration_percent": 54.4,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [75.0, 150.0],
#             "eta_stages_percent": [31.1, 32.4],
#             "eta_th_stages_percent": [30.8, 31.2],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 2 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [75.0, 150.0],
#             "eta_stages_percent": [33.0, 33.2],
#             "eta_th_stages_percent": [30.0, 30.5],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 2 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_300kw():
#     """Init vals for a Biogas plant with 300 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 75.0,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 75.0,
#         },
#     }


# def params_2g_320kw():
#     """Params for a Biogas plant with 320 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 2050,
#         "cap_gas_m3": 855,
#         "gas_fill_min_percent": 2.0,
#         "gas_fill_max_percent": 98.0,
#         "ch4_concentration_percent": 63.9,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [40.0, 80.0, 120.0, 160.0],
#             "eta_stages_percent": [31.1, 32.2, 33.3, 34.4],
#             "eta_th_stages_percent": [30.8, 31.2, 32.3, 32.3],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [85.0, 110.0, 135.0, 160.0],
#             "eta_stages_percent": [32.2, 33.0, 33.8, 34.6],
#             "eta_th_stages_percent": [30.0, 30.5, 31.0, 31.5],
#             "restarts_per_day": 6,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_320kw():
#     """Init vals for a Biogas plant with 320 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 80.0,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 85.0,
#         },
#     }


# def params_2g_2050kw():
#     """Params for a Biogas plant with 2050 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 14_500,
#         "cap_gas_m3": 13_000,
#         "gas_fill_min_percent": 5,
#         "gas_fill_max_percent": 95,
#         "ch4_concentration_percent": 50.0,
#         "has_heat_storage": False,
#         "num_chps": 2,
#         "chp0": {
#             "pn_stages_kw": [700, 1150],
#             "eta_stages_percent": [37.0, 44.0],
#             "eta_th_stages_percent": [40.0, 44.0],
#             "restarts_per_day": 4,
#             "active_min_s": 0,
#             "active_max_s_per_day": 16 * 3_600,
#             "inactive_min_s": 2 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [600, 900],
#             "eta_stages_percent": [37, 40.0],
#             "eta_th_stages_percent": [42.0, 43.0],
#             "restarts_per_day": 4,
#             "active_min_s": 0,
#             "active_max_s_per_day": 16 * 3_600,
#             "inactive_min_s": 2 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_2g_2050kw():
#     """Init vals for a Biogas plant with 2050 kw nominal power and
#     two CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 700,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 600,
#         },
#     }


# def params_3g_555kw():
#     """Params for a Biogas plant with 555 kw nominal power and
#     three CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 6000,
#         "cap_gas_m3": 2500,
#         "gas_fill_min_percent": 5,
#         "gas_fill_max_percent": 95,
#         "ch4_concentration_percent": 51.6,
#         "has_heat_storage": False,
#         "num_chps": 3,
#         "chp0": {
#             "pn_stages_kw": [125, 250],
#             "eta_stages_percent": [31.2, 34.7],
#             "eta_th_stages_percent": [42.0, 45.9],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 18 * 3_600,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [125, 250],
#             "eta_stages_percent": [34.3, 36.1],
#             "eta_th_stages_percent": [41.9, 43],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 18 * 3_600,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp2": {
#             "pn_stages_kw": [27.5, 55],
#             "eta_stages_percent": [28.6, 30],
#             "eta_th_stages_percent": [35.1, 36.1],
#             "restarts_per_day": 0,
#             "active_min_s": 2 * 3_600,
#             "active_max_s_per_day": 18 * 3_600,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_3g_555kw():
#     """Init vals for a Biogas plant with 555 kw nominal power and
#     three CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 250,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 125,
#         },
#         "chp2": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 55,
#         },
#     }


# def params_3g_1500kw():
#     """Params for a Biogas plant with 1500 kw nominal power and
#     three CHP CNGs.
#     """
#     return {
#         "gas_m3_per_day": 16_040,
#         "cap_gas_m3": 25_200,
#         "gas_fill_min_percent": 5,
#         "gas_fill_max_percent": 95,
#         "ch4_concentration_percent": 53.9,
#         "has_heat_storage": False,
#         "num_chps": 3,
#         "chp0": {
#             "pn_stages_kw": [125.0, 250.0, 375.0, 500.0, 625.0, 750.0],
#             "eta_stages_percent": [26.1, 27.0, 27.3, 28.0, 27.7, 27.5],
#             "eta_th_stages_percent": [50.0, 50.2, 50.5, 50.9, 51.0, 52.1],
#             "restarts_per_day": 0,
#             "active_min_s": 3 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp1": {
#             "pn_stages_kw": [330.0, 440.0, 550],
#             "eta_stages_percent": [27.9, 28.5, 29.1],
#             "eta_th_stages_percent": [45.9, 47.3, 51.0],
#             "restarts_per_day": 0,
#             "active_min_s": 1 * 3_600,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 0,
#             "inactive_max_s_per_day": 0,
#         },
#         "chp2": {
#             "pn_stages_kw": [100, 200],
#             "eta_stages_percent": [28.8, 29.9],
#             "eta_th_stages_percent": [29.7, 30.1],
#             "restarts_per_day": 0,
#             "active_min_s": 0,
#             "active_max_s_per_day": 0,
#             "inactive_min_s": 1 * 3_600,
#             "inactive_max_s_per_day": 0,
#         },
#     }


# def inits_3g_1500kw():
#     """Init vals for a Biogas plant with 1500 kw nominal power and
#     three CHP CNGs.
#     """
#     return {
#         "gas_fill_percent": 50,
#         "chp0": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 125.0,
#         },
#         "chp1": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 330,
#         },
#         "chp2": {
#             "active_s": 0,
#             "active_s_per_day": 0,
#             "inactive_s": 0,
#             "inactive_s_per_day": 0,
#             "restarts": 0,
#             "p_kw": 100,
#         },
#     }



"""This module contains the state model for pv."""

from typing import cast

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class PVState(ModelState):
    """State parameters of pv model.

    See :class:`pysimmods.model.state.ModelState` for additional
    information.

    Parameters
    ----------
    inits : dict
        Contains the initial configuration of this PV plant. See
        attributes section for specific to the PV plant.

    Attributes
    ----------
    t_module_deg_celsius : float
        Temperature of the module in [°C].

    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)
        self.t_module_deg_celsius: float = cast(
            "float", inits["t_module_deg_celsius"]
        )



"""This module contains the input model for the chp."""

from pysimmods.base.inputs import ModelInputs


class CHPCNGInputs(ModelInputs):
    """A CHP CNG input class.

    Attributes
    ----------
    gas_in_m3: float
        Available gas (e.g., from a gas storage) in [m^3].
    gas_critical: bool
        If set to True, this unit will reduce generation.
    heat_critical: bool
        If set to True, this unit will increase generation. Has lower
        priority than gas_critical
    """

    def __init__(self):
        super().__init__()

        self.gas_in_m3: float = 0.0
        self.gas_critical = False
        self.heat_critical = False

    def reset(self):
        """To be called at the end of each step."""
        # for attr in self.__dict__.keys():
        #     setattr(self, attr, None)
        self.gas_in_m3 = 0.0
        self.gas_critical = False
        self.heat_critical = False



"""This module provides configurations for CHP CNGs.

The CHP Compressed Natural Gas are used by the Biogas model.
The configurations are derived from ... (TODO).

"""

from pathlib import Path
from typing import Any, cast

import numpy as np
import pandas as pd

from pysimmods.base.types import ModelInitVals, ModelParams

BASE_DATA = (Path(__file__) / ".." / "base_data.csv").resolve()


def chpcng_preset(
    pn_max_kw: float,
    randomize_params: bool = False,
    randomize_inits: bool = False,
    seed: int | None = None,
    **kwargs: dict[str, Any],
) -> tuple[ModelParams, ModelInitVals]:
    rng = np.random.default_rng(seed)
    all_params = _read_all_params()

    if randomize_params:
        # "Magic values" derived from base_data.csv
        active_max = max(0, int(rng.integers(18) - 12))
        if active_max > 0:
            active_max = (active_max + 12) * 3600
        params: ModelParams = {
            "restarts_per_day": max(0, int(rng.integers(-6, 6))),
            "active_min_s": max(0, int(rng.integers(-3, 3))) * 3600,
            "active_max_s_per_day": active_max,
            "inactive_min_s": max(0, int(rng.integers(-2, 2))) * 3600,
            "inactive_max_s_per_day": 0,
        }
        n_stages = int(rng.integers(2, 7))
        min_ratio = float(rng.uniform(0.3, 0.6))
        pn_min_kw = pn_max_kw * min_ratio
        step_size = (pn_max_kw - pn_min_kw) / (n_stages - 1)
        eta_min = float(rng.uniform(25.0, 40.0))
        eta_max = float(rng.uniform(eta_min, max(eta_min + 2, 45.0)))
        eta_step_size = (eta_max - eta_min) / (n_stages - 1)
        eta_th_min = float(rng.uniform(28.0, 50.0))
        eta_th_max = float(rng.uniform(eta_th_min, max(eta_th_min + 2, 55.0)))
        eta_th_step_size = (eta_th_max - eta_th_min) / (n_stages - 1)

        pn_stages: list[float] = []
        eta_stages: list[float] = []
        eta_th_stages: list[float] = []
        for i in range(n_stages):
            pn_stages.append(pn_min_kw + i * step_size)
            eta_stages.append(
                eta_min + i * eta_step_size * rng.uniform(0.5, 2.5)
            )
            eta_th_stages.append(
                eta_th_min + i * eta_th_step_size * rng.uniform(0.5, 2.5)
            )
        params["pn_stages_kw"] = pn_stages
        params["eta_stages_percent"] = eta_stages
        params["eta_th_stages_percent"] = eta_th_stages

    else:
        closest_name = ""
        closest_distance = 2**32 - 1
        for n, params in all_params.items():
            dist = abs(
                pn_max_kw - cast("list[float]", params["pn_stages_kw"])[-1]
            )

            if dist < closest_distance:
                closest_distance = dist
                closest_name = n
        params = all_params[closest_name]

    return params, _build_inits(
        params, randomize_inits, int(rng.integers(2**32 - 1))
    )


def chpcng_preset_by_name(
    name: str, randomize_inits: bool = False, seed: int | None = None, **kwargs
) -> tuple[ModelParams, ModelInitVals]:
    params = _read_all_params()[name]

    return params, _build_inits(params, randomize_inits, seed)


def _read_all_params() -> dict[str, ModelParams]:
    data = pd.read_csv(BASE_DATA)

    all_params = {}

    for _, row in data.iterrows():
        params = {
            "restarts_per_day": cast("int", row["restarts_per_day"]),
            "active_min_s": cast("int", row["active_min_s"]),
            "active_max_s_per_day": cast("int", row["active_max_s_per_day"]),
            "inactive_min_s": cast("int", row["inactive_min_s"]),
            "inactive_max_s_per_day": cast(
                "int", row["inactive_max_s_per_day"]
            ),
            "pn_stages_kw": [],
            "eta_stages_percent": [],
            "eta_th_stages_percent": [],
        }
        for j in range(6):
            pn_kw = cast("float", row[f"pn_{j:02d}"])
            if pd.isnull(pn_kw):
                break
            params["pn_stages_kw"].append(pn_kw)
            params["eta_stages_percent"].append(
                cast("float", row[f"eta_{j:02d}"])
            )
            params["eta_th_stages_percent"].append(
                cast("float", row[f"eta_th_{j:02d}"])
            )
        all_params[row["name"]] = params

    return all_params


def _build_inits(
    params: ModelParams, randomize_inits: bool = False, seed: int | None = None
) -> ModelInitVals:
    rng = np.random.default_rng(seed)
    if randomize_inits:
        inits: ModelInitVals = {
            "active_s": max(
                0, int(rng.integers(-1, cast("int", params["active_min_s"])))
            ),
            "active_s_per_day": max(
                0,
                int(
                    rng.integers(
                        -1, cast("int", params["active_max_s_per_day"])
                    )
                ),
            ),
            "inactive_s": max(
                0, int(rng.integers(-1, cast("int", params["inactive_min_s"])))
            ),
            "inactive_s_per_day": 0,
            "restarts": max(
                0,
                int(rng.integers(-1, cast("int", params["restarts_per_day"]))),
            ),
            "p_kw": float(
                rng.choice(cast("list[float]", params["pn_stages_kw"]))
            ),
        }
    else:
        inits: ModelInitVals = {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": cast("list[float]", params["pn_stages_kw"])[0],
        }

    return inits



"""This module contains the config of the Wind Turbine System."""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams
from pysimmods.generator.windsim.config import WindPowerPlantConfig
from pysimmods.other.invertersim.config import InverterConfig


class WindSystemConfig(ModelConfig):
    def __init__(self, params: ModelParams):
        super().__init__(params)
        cast("ModelParams", params["wind"])["sign_convention"] = (
            self.sign_convention
        )
        cast("ModelParams", params["inverter"])["sign_convention"] = (
            self.sign_convention
        )

        # Those are duplicate to the original configs
        self._turbine: WindPowerPlantConfig = WindPowerPlantConfig(
            cast("ModelParams", params["wind"])
        )
        self._inverter: InverterConfig = InverterConfig(
            cast("ModelParams", params["inverter"])
        )

        self.s_max_kva = self._inverter.s_max_kva
        self.q_control = self._inverter.q_control
        self.cos_phi = self._inverter.cos_phi
        self.p_max_kw = self._turbine.p_max_kw
        self.p_min_kw = self._turbine.p_min_kw

        self.default_p_schedule = None
        self.default_q_schedule = None



from pysimmods.base.types import ModelInitVals, ModelParams


def windsys_preset(
    pn_max_kw: float,
    wind_profile: str = "hellmann",
    temperature_profile: str = "linear_gradient",
    air_density_profile: str = "barometric",
    power_method: str = "power_curve",
    turbine_type: str | None = None,
    cos_phi: float = 0.9,
    q_control: str = "prioritize_p",
    inverter_mode: str = "inductive",
    randomize_params: bool = False,
    seed: int | None = None,
    **kwargs,
) -> tuple[ModelParams, ModelInitVals]:
    from pysimmods.generator.windsim.presets import wind_preset

    wparams, winit = wind_preset(
        pn_max_kw,
        wind_profile,
        temperature_profile,
        air_density_profile,
        power_method,
        turbine_type,
        randomize_params,
        seed,
        **kwargs,
    )

    params: ModelParams = {
        "wind": wparams,
        "inverter": {
            "sn_kva": pn_max_kw / cos_phi,
            "q_control": q_control,
            "cos_phi": cos_phi,
            "inverter_mode": inverter_mode,
        },
        "sign_convention": "active",
    }

    inits: ModelInitVals = {"wind": winit, "inverter": {}}

    return params, inits



"""This module contains the config model of the PV plant system."""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams
from pysimmods.generator.pvsim.config import PVConfig
from pysimmods.other.invertersim.config import InverterConfig


class PVSystemConfig(ModelConfig):
    """Config parameters of the PV plant System.

    Consists of a :class:`~.PVConfig` and an :class:`~.InverterConfig`
    object.

    """

    def __init__(self, params: ModelParams):
        super().__init__(params)
        pv_params = cast("ModelParams", params["pv"])
        inv_params = cast("ModelParams", params["inverter"])
        pv_params["sign_convention"] = self.sign_convention
        inv_params["sign_convention"] = self.sign_convention

        self.pv = PVConfig(pv_params)
        self.inverter = InverterConfig(inv_params)
        self.default_p_schedule = None
        self.default_q_schedule = None

        self.p_max_kw = self.pv.p_max_kw
        self.p_min_kw = self.pv.p_min_kw
        self.s_max_kva = self.inverter.s_max_kva

    @property
    def q_control(self):
        return self.inverter.q_control

    @property
    def cos_phi(self):
        return self.inverter.cos_phi



import os

import matplotlib.pyplot as plt
import numpy as np
from natsort import natsorted


def analyze_power(
    data, step_size, report_file, output_path, full_report, config
):
    ef = step_size / 3_600
    name = config["name"]
    report_file.append(f"## {config['topic']} Analysis\n")
    p_stats = dict()
    q_stats = dict()
    data = data[natsorted(data.columns)]
    for colo in data.columns:
        col = colo.replace("___", ".").replace("__", "-")
        _, key, attr = col.split(".")
        if full_report:
            _plot_series(data[colo], ef, f"{key}.{attr}", name, output_path)
        stats = {
            "Min": data[colo].min(),
            "Max": data[colo].max(),
            "Mean": data[colo].mean(),
            "Std": data[colo].std(),
            "Sum": data[colo].sum(),
            "Energy": data[colo].sum() * ef,
        }
        if attr == "p_mw":
            p_stats[key] = stats
        else:
            q_stats[key] = stats

    # Aggregated analyis
    total_p = data[[col for col in data.columns if "p_mw" in col]].sum(axis=1)
    pfile = _plot_series(
        total_p, ef, f"{config['total_name']}.p_mw", name, output_path
    )
    total_q = data[[col for col in data.columns if "q_mvar" in col]].sum(
        axis=1
    )
    qfile = _plot_series(
        total_q, ef, f"{config['total_name']}.q_mvar", name, output_path
    )
    p_stats["total"] = {
        "Min": total_p.min(),
        "Max": total_p.max(),
        "Mean": total_p.mean(),
        "Std": total_p.std(),
        "Sum": total_p.sum(),
        "Energy": total_p.sum() * ef,
    }
    q_stats["total"] = {
        "Min": total_q.min(),
        "Max": total_q.max(),
        "Mean": total_q.mean(),
        "Std": total_q.std(),
        "Sum": total_q.sum(),
        "Energy": total_q.sum() * ef,
    }

    _create_report(p_stats, "MW", config["topic"], report_file)
    report_file.append("")
    _create_report(q_stats, "MVAr", config["topic"], report_file)

    report_file.append(
        f"\n![Total_{config['topic']}_P]({pfile})" + "{width=100%}\n"
    )
    report_file.append(
        f"\n![Total_{config['topic']}_Q]({qfile})" + "{width=100%}\n"
    )

    total_s = np.sqrt(
        np.square(total_p.sum() * ef) + np.square(total_q.sum() * ef)
    )
    return (total_p, total_q, total_s, np.ceil(np.max(total_p)))


def _plot_series(data, ef, col, name, output_path):
    series = data.values
    key, attr = col.split(".")
    if attr == "p_mw":
        active = "active"
        unit = "MW"
    else:
        active = "reactive"
        unit = "MVAr"

    annual = np.sort(series)[::-1]

    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(12, 12))
    ax1.plot(series)
    ax1.set_title(f"{key} {active} power")
    # ax1.set_xlabel("time (15 minute steps")
    ax1.set_ylabel(unit)

    ax2.plot(annual)
    ax2.set_title(f"{key} annual curve {active} power")
    ax2.set_ylabel(unit)

    ax3.plot(np.cumsum(ef * series))
    ax3.set_title(f"{key} cummulated {active} power")
    ax3.set_xlabel("time (15 minute steps")
    ax3.set_ylabel(unit)

    filename = os.path.join(output_path, f"{name}_{key}_{active}_power.png")
    plt.savefig(filename, dpi=300, bbox_inches="tight")
    plt.close()
    return filename


def _create_report(stats, unit, topic, report_file):
    report_file.append(
        f"| {topic} | Min ({unit}) | Max ({unit}) | Mean ({unit}) |"
        f" StD ({unit}) | Sum ({unit}) | Energy ({unit}h) |"
    )
    report_file.append(
        "|:---------|---------:|---------:|----------:|"
        "---------:|---------:|-------------:|"
    )
    for key, vals in stats.items():
        report_file.append(
            f"| {key} | {vals['Min']:.3f} "
            f"| {vals['Max']:.3f} "
            f"| {vals['Mean']:.3f} "
            f"| {vals['Std']:.3f} "
            f"| {vals['Sum']:.3f} "
            f"| {vals['Energy']:.3f} |"
        )

    report_file.append("")



"""This module contains the base class for all model inputs."""

from datetime import datetime

from pysimmods.util.date_util import convert_dt


class ModelInputs:
    """Base class for model inputs.

    All attributes should be resetted after each step in the model
    implementations. Providing None is usually a valid input (though
    not in all cases).

    Attributes
    ----------
    p_set_kw : float
        Target electrical activate power in [kW].
    q_set_kvar : float
        Target electrical reactive power in [kVAr].
    step_size : int
        Step size for the net step in [s].
    now_dt: datetime | str | int
        Current timestamp as datetime object, ISO datestring or unix
        timestamp.

    """

    def __init__(self) -> None:
        self.p_set_kw: float | None = None
        self.q_set_kvar: float | None = None
        self.step_size: int = 1
        self._now_dt: float | None = None

    def reset(self) -> None:
        """To be called at the end of each step."""
        for attr in self.__dict__.keys():
            setattr(self, attr, None)

    @property
    def now_dt(self) -> datetime:
        """The current date and time of the model"""
        return self._now_dt

    @now_dt.setter
    def now_dt(self, now: datetime | str | int) -> None:
        self._now_dt = convert_dt(now)



from typing import cast

from pysimmods.base.model import ModelState
from pysimmods.base.types import ModelInitVals


class HeatStorageState(ModelState):
    """Captures the state of the heat storage"""

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.t_c: float = cast("float", inits["storage_t_c"])
        self.t_chilled: float = 0.0
        self.e_th_in_max_kwh: float = 0.0
        self.e_th_in_min_kwh: float = 0.0



"""This module contains an inverter model"""

from copy import copy
from math import sqrt

from pysimmods.base.qgenerator import QGenerator
from pysimmods.base.types import ModelInitVals, ModelParams, QControl
from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.other.invertersim.inputs import InverterInputs
from pysimmods.other.invertersim.state import InverterState


class Inverter(QGenerator[InverterConfig, InverterState, InverterInputs]):
    """Inverter model

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict, optional
        A *dict* containing initialization parameters.

    Attributes
    ----------
    config : :class:`~.InverterConfig`
        Configuration parameters of the inverter.
    state : :class:`~.InverterState`
        Initialization parameters of the inverter.
    inputs : :class:`~.InverterInputs`
        Input parameters of the inverter.

    """

    def __init__(
        self, params: ModelParams, inits: ModelInitVals | None = None
    ):
        if inits is None:
            inits = {}

        self.config: InverterConfig = InverterConfig(params)
        self.state: InverterState = InverterState(inits)
        self.inputs: InverterInputs = InverterInputs()

    def step(self, pretend=False) -> InverterState:
        """Perform simulation step"""
        self._check_inputs()

        next_state = copy(self.state)
        next_state._inductive = self.inputs._inductive
        self._calculate(next_state)

        if not pretend:
            self.state = next_state

        self.inputs.reset()

        return next_state

    def _check_inputs(self):
        if self.inputs.p_set_kw is None:
            self.inputs.p_set_kw = self.inputs.p_in_kw

        if self.inputs.cos_phi_set is None:
            self.inputs.cos_phi_set = self.config.cos_phi

        if self.inputs._inductive is None:
            self.inputs._inductive = self.config.inverter_mode == "inductive"

    def _calculate(self, next_state: InverterState):
        # Can't output more power than is available
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for apparent power
        s_kva = p_kw / self.inputs.cos_phi_set
        s_kva = min(s_kva, self.config.s_max_kva)

        # Use updated apparent power to calculate active power
        p_max_kw = s_kva * self.inputs.cos_phi_set
        p_kw = min(p_kw, p_max_kw)

        # Calculate maximum reactive power
        q_max_kvar = sqrt(self.config.s_max_kva**2 - p_kw**2)

        # Check if reactive power setpoint is provided
        if self.inputs.q_set_kvar is None:
            # No; use "remaining" current apparent power for reactive
            # power, preserving the cos phi
            q_kvar = sqrt(s_kva**2 - p_kw**2)
            q_kvar = min(q_max_kvar, q_kvar)
            if next_state._inductive:
                q_kvar *= -1
        else:
            # Yes; use "remaining" maximum apparent power for reactive
            # power, the cos phi may change
            q_set_sign = 1 if self.inputs.q_set_kvar >= 0 else -1
            q_kvar = min(
                abs(self.inputs.q_set_kvar), abs(self.config.s_max_kva)
            )
            s_kva = sqrt(p_kw**2 + q_kvar**2)

            # Limit active or reactive power according to the q-control
            # mode
            if s_kva > self.config.s_max_kva:
                if self.config.q_control == QControl.PRIORITIZE_P:
                    q_kvar = q_max_kvar
                elif self.config.q_control == QControl.PRIORITIZE_Q:
                    p_kw = sqrt(self.config.s_max_kva**2 - q_kvar**2)

            q_kvar *= q_set_sign

        if s_kva > 0:
            next_state.cos_phi = p_kw / s_kva
        else:
            next_state.cos_phi = self.inputs.cos_phi_set
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar



"""This module contains the heat storage model that is used by the
CHP LPG.
"""

import math
from copy import deepcopy

from pysimmods.base.buffer import Buffer
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.other.heatstoragesim.config import HeatStorageConfig
from pysimmods.other.heatstoragesim.inputs import HeatStorageInputs
from pysimmods.other.heatstoragesim.state import HeatStorageState

C_WATER = 4.18
"""Specific heat capacity of water in [kJ/(kg*K)]"""


class HeatStorage(
    Buffer[HeatStorageConfig, HeatStorageState, HeatStorageInputs]
):
    """Model of a heat storage.

    Parameters
    ----------
    params : dict
        A *dict* containing values for all attributes. The keys all
        have the prefix "storage_".
    inits : dict
        A *dict* containing state variables for the storage. The keys
        all have the prefix "storage_".

    Attributes
    ----------
    cap_l : float
        Capacity of the buffer storage in [l].
    consumption_kwh_per_day : float
        Average loss of energy in [kWh/day].
    t_min_c : float
        Minimal water temperature of the buffer storage in [°C].
    t_max_c : float
        Maximal water temperature of the buffer storage in [°C].
    t_c : float
        Current water temperature of the buffer storage in [°C].
    env_t_c : float, optional
        Current temperature of the environment in [°C]. The storage
        is assumed to be placed indoor and, therefore, defaults to
        19.0 °C.

    """

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = HeatStorageConfig(params)
        self.state = HeatStorageState(inits)
        self.inputs = HeatStorageInputs()

    def step(self, pretend: bool = False) -> HeatStorageState:
        next_state = deepcopy(self.state)

        next_state.e_th_in_max_kwh = self.get_absorbable_energy()
        next_state.e_th_in_min_kwh = self.get_available_energy()

        e_th_delta_kj = (
            self.inputs.e_th_prod_kwh + self.inputs.e_th_demand_kwh
        ) * 3_600

        next_state.t_chilled = self._calculate_chilled_t()
        t_new = (
            e_th_delta_kj / (C_WATER * self.config.cap_l)
            + next_state.t_chilled
        )

        next_state.t_c = max(t_new, self.config.env_t_c)

        if not pretend:
            self.state = next_state
        return next_state

    def _calculate_chilled_t(self):
        """Calculate the chilled temperature.

        This function requires that the :attr:`.env_t_c` is set.

        Returns
        -------
        float
            The chilled temperature

        """

        chill = (
            self.config.consumption_kwh_per_day
            * 3_600_00
            / (C_WATER * 1_000 * self.config.cap_l * 24)
        )
        t_next = self.state.t_c - chill

        # Prevent math domain/division by zero errors
        dif_t_next_env = max(t_next - self.config.env_t_c, 1e-3)
        dif_t_init_env = max(self.state.t_c - self.config.env_t_c, 1e-3)
        div_t = max(dif_t_next_env / dif_t_init_env, 1e-3)

        chill_coef = math.log(div_t) / t_next
        t_chilled = (self.state.t_c - self.config.env_t_c) * math.pow(
            math.e, chill_coef
        ) + self.config.env_t_c

        return t_chilled

    def get_absorbable_energy(self) -> float:
        """Return the amount of energy the storage can absorb."""
        e_th_in_max_kwh = (
            C_WATER
            * self.config.cap_l
            * (self.config.t_max_c - self.state.t_c)
            / 3_600
        )

        return e_th_in_max_kwh

    def get_available_energy(self) -> float:
        """Return the amount of energy the storage has absorbed."""

        e_th_in_min_kwh = (
            C_WATER
            * self.config.cap_l
            * max(0, self.state.t_c - self.config.t_min_c)
            / 3_600
        )

        return e_th_in_min_kwh

    def absorb_energy(self, e_th_prod, e_th_demand):
        # 1 kWh = 3600 kJ
        # One and only one of prod and demand has to be negative
        e_th_delta_kj = (e_th_prod + e_th_demand) * 3_600

        t_chilled = self._calculate_chilled_t()
        t_new = e_th_delta_kj / (C_WATER * self.config.cap_l) + t_chilled

        self.state.t_c = max(t_new, self.config.env_t_c)



"""
This module contains a model of a chp system with a chp and a
household. The household has a heat demand which allows the chp
to run for at least 5000 hours in a year.

"""

import logging
from copy import deepcopy
from typing import cast

from pysimmods.base.generator import Generator
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.generator.chplpgsim.chplpg import CHPLPG
from pysimmods.generator.chplpgsystemsim.config import CHPLPGSystemConfig
from pysimmods.generator.chplpgsystemsim.inputs import CHPLPGSystemInputs
from pysimmods.generator.chplpgsystemsim.state import CHPLPGSystemState
from pysimmods.other import HeatDemand, heatdemand_preset

LOG = logging.getLogger(__name__)


class CHPLPGSystem(
    Generator[CHPLPGSystemConfig, CHPLPGSystemState, CHPLPGSystemInputs]
):
    """CHP system with CHP and household"""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        # First, create the chp
        chp_params = cast("ModelParams", params["chp"])
        chp_params["sign_convention"] = params.get(
            "sign_convention", "passive"
        )
        self.chp = CHPLPG(chp_params, cast("ModelInitVals", inits["chp"]))

        # Get the thermal capabilities of the chp
        p_th_min_kw = (
            self.chp.config.p_min_kw * self.chp.config.p_2_p_th_percent * 0.01
        )

        # Second, create an appropriate household
        # FIXME: Seeding is non-deterministic here!
        self.heatdemand = HeatDemand(
            *heatdemand_preset(-p_th_min_kw, seed=None)
        )

        self.config = CHPLPGSystemConfig(params)
        self.state = CHPLPGSystemState(inits)
        self.inputs = CHPLPGSystemInputs()

    def step(self, pretend: bool = False) -> CHPLPGSystemState:
        """Perform a simulation step"""
        next_state = deepcopy(self.state)

        err = None
        try:
            # First step the household
            self.heatdemand.inputs.now_dt = self.inputs.now_dt
            self.heatdemand.inputs.step_size = self.inputs.step_size
            self.heatdemand.inputs.day_avg_t_air_deg_celsius = (
                self.inputs.day_avg_t_air_deg_celsius
            )
            self.heatdemand.step()
        except TypeError as te:
            err = te

        # Second step the chp
        p_set_kw = self._check_setpoint()
        self.chp.set_now_dt(self.inputs.now_dt)
        if self.inputs.e_th_demand_set_kwh is not None:
            self.chp.inputs.e_th_demand_set_kwh = (
                self.inputs.e_th_demand_set_kwh * self.config.e_th_demand_sign
            )
        elif err is None:
            self.chp.inputs.e_th_demand_set_kwh = (
                self.heatdemand.state.e_th_kwh
            )
        else:
            msg = (
                "One input is missing for CHP. Must at least provide "
                "one of 'day_avg_t_air_deg_celsius' or "
                "'e_th_demand_kwh'!"
            )
            LOG.error(msg)
            raise ValueError(msg)

        self.chp.set_step_size(self.inputs.step_size)
        self.chp.set_p_kw(p_set_kw * self.config.gsign)
        nstate_chp = self.chp.step(pretend)

        # Update the references
        next_state.p_kw = nstate_chp.p_kw
        next_state.q_kvar = nstate_chp.q_kvar
        next_state.p_th_kw = nstate_chp.p_th_kw
        next_state.storage_t_c = nstate_chp.storage_t_c
        next_state.lubricant_l = nstate_chp.lubricant_l
        if self.inputs.e_th_demand_set_kwh is not None:
            next_state.e_th_demand_kwh = self.inputs.e_th_demand_set_kwh
        else:
            next_state.e_th_demand_kwh = self.heatdemand.state.e_th_kwh

        if not pretend:
            self.state = next_state
        return next_state

    def _check_setpoint(self):
        setpoint = self.inputs.p_set_kw
        if setpoint is not None:
            return abs(setpoint)
        hour = self.inputs.now_dt.hour
        default = self.config.default_schedule[hour]
        setpoint = self.config.p_max_kw * default / 100

        return abs(setpoint)

    def get_state(self):
        """Get state"""
        state_dict = {
            "household": self.heatdemand.get_state(),
            "chp": self.chp.get_state(),
        }
        return state_dict

    def set_state(self, state):
        """Set state"""
        self.heatdemand.set_state(state["household"])
        self.chp.set_state(state["chp"])

        self.state.p_kw = self.chp.state.p_kw
        self.state.q_kvar = self.chp.state.q_kvar
        self.state.p_th_kw = self.chp.state.p_th_kw
        self.state.storage_t_c = self.chp.state.storage_t_c
        self.state.lubricant_l = self.chp.state.lubricant_l



"""This module contains different DummyModels, which only implement the
the model interfaces. Their purpose is mainly for testing.

"""

from copy import deepcopy
from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.inputs import ModelInputs
from pysimmods.base.model import Model
from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals, ModelParams


class DummyConfig(ModelConfig):
    def __init__(self, params: ModelParams):
        super().__init__(params)

        self.p_max_kw: float = cast("float", params.get("p_max_kw", 500.0))
        self.p_min_kw: float = cast("float", params.get("p_min_kw", 250.0))
        self.q_min_kvar: float = cast(
            "float", params.get("q_min_kvar", -550.0)
        )
        self.q_max_kvar: float = cast("float", params.get("q_max_kvar", 550.0))
        self.default_p_schedule = [375.0] * 24
        self.default_q_schedule = [125.0] * 24


class DummyState(ModelState):
    pass


class DummyInputs(ModelInputs):
    pass


class DummyModel(Model[DummyConfig, DummyState, DummyInputs]):
    """The dummy base model."""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = DummyConfig(params)
        self.state = DummyState(inits)
        self.inputs = DummyInputs()

    def step(self, pretend: bool = False) -> DummyState:
        next_state = deepcopy(self.state)

        # Check inputs
        next_state.p_kw = (
            0.0 if self.inputs.p_set_kw is None else self.inputs.p_set_kw
        )
        next_state.q_kvar = (
            0.0 if self.inputs.q_set_kvar is None else self.inputs.q_set_kvar
        )

        # Apply constraints
        next_state.p_kw = min(
            self.config.p_max_kw, max(self.config.p_min_kw, next_state.p_kw)
        )
        next_state.q_kvar = min(
            self.config.q_max_kvar,
            max(self.config.q_min_kvar, next_state.q_kvar),
        )

        if not pretend:
            self.state = next_state
        return next_state

    def get_pn_max_kw(self):
        return self.config.p_max_kw

    def get_pn_min_kw(self):
        return self.config.p_min_kw

    def get_qn_max_kvar(self):
        return self.config.q_max_kvar

    def get_qn_min_kvar(self):
        return self.config.q_min_kvar

    def set_p_kw(self, p_kw: float) -> None:
        self.inputs.p_set_kw = p_kw

    def get_p_kw(self):
        return self.state.p_kw

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar

    def get_q_kvar(self):
        return self.state.q_kvar



"""
This module contains a household class that simulates thermal power demand
for the chp.

"""

import math
from copy import deepcopy
from datetime import timedelta

from pysimmods.base.consumer import Consumer
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.other.heatdemandsim.config import HeatDemandConfig
from pysimmods.other.heatdemandsim.inputs import HeatDemandInputs
from pysimmods.other.heatdemandsim.state import HeatDemandState


class HeatDemand(
    Consumer[HeatDemandConfig, HeatDemandState, HeatDemandInputs]
):
    """A simple model for thermal power demand."""

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = HeatDemandConfig(params)
        self.state = HeatDemandState(inits)
        self.inputs = HeatDemandInputs()

    def step(self, pretend: bool = False) -> HeatDemandState:
        """Perform a simulation step"""

        next_state = deepcopy(self.state)

        # calculate allocation temperature for the current day
        # (weighted mean temperature of the current and the last
        # three days)
        t_alloc = (
            1.0 * self.inputs.day_avg_t_air_deg_celsius
            + 0.5 * self.state.t_last_1_deg_celsius
            + 0.25 * self.state.t_last_2_deg_celsius
            + 0.125 * self.state.t_last_3_deg_celsius
        ) / (1 + 0.5 + 0.25 + 0.125)

        if t_alloc >= self.config.const_v_0:
            t_alloc = self.config.const_v_0 * 0.99

        # profile function 'SigLinDe': calculate daily consumption
        # depending on the allocation temperature. Has both a sigmoid
        # and a linear part
        tmp = self._daily_consumption_pow(t_alloc)
        sigmoid = self.config.const_a / (1.0 + tmp) + self.config.const_d
        linear = max(
            self.config.const_m_h * t_alloc + self.config.const_b_h,
            self.config.const_m_w * t_alloc + self.config.const_b_w,
        )

        daily_cons = (
            self.config.consumer_const
            * (sigmoid + linear)
            * self.config.weekday_const[self.inputs.now_dt.weekday()]
        )

        next_state.e_th_kwh = float(
            daily_cons
            * self.config.degree_of_efficiency
            * self.config.load_profile[self.inputs.now_dt.hour]
            * self.inputs.step_size
            / 3_600
        )

        next_dt = self.inputs.now_dt + timedelta(seconds=self.inputs.step_size)
        if next_dt.day > self.inputs.now_dt.day:
            # we reached a new day
            next_state.t_last_3_deg_celsius = next_state.t_last_2_deg_celsius
            next_state.t_last_2_deg_celsius = next_state.t_last_1_deg_celsius
            next_state.t_last_1_deg_celsius = (
                self.inputs.day_avg_t_air_deg_celsius
            )

        if not pretend:
            self.state = next_state
        return next_state

    def _daily_consumption_pow(self, t_alloc):
        return math.pow(
            self.config.const_b / (t_alloc - self.config.const_v_0),
            self.config.const_c,
        )

    @property
    def e_th_kwh(self):
        """Returns the current thermal power demand"""
        return self.state.e_th_kwh

    def get_state(self) -> dict:
        """Return the current state of the model.

        Returns
        -------
        dict
            The current state of the model in form of a dictionary
            containing entries for all state variables. Returned dict
            can be assigned to the *inits* argument when creating a new
            model instance.

        """

        try:
            return {
                attr: getattr(self.state, attr)
                for attr in self.state.__slots__
            }
        except AttributeError:
            return deepcopy(self.state.__dict__)

    def set_state(self, state: dict) -> None:
        """Set the current state of the model.

        Parameters
        ----------
        state : dict
            A *dict* containing entries for all state variables.

        """
        for attr, value in state.items():
            setattr(self.state, attr, value)



from pysimmods.generator.biogassim.biogas import BiogasPlant as BiogasPlant
from pysimmods.generator.biogassim.presets import (
    biogas_preset as biogas_preset,
)
from pysimmods.generator.chplpgsystemsim.chplpg_system import (
    CHPLPGSystem as CHPLPGSystem,
)
from pysimmods.generator.chplpgsystemsim.presets import (
    chp_preset as chp_preset,
)
from pysimmods.generator.pvsystemsim.presets import pv_preset as pv_preset
from pysimmods.generator.pvsystemsim.pvpsystem import (
    PVPlantSystem as PVPlantSystem,
)
from pysimmods.generator.windsystemsim.presets import (
    windsys_preset as windsys_preset,
)
from pysimmods.generator.windsystemsim.wind_turbine_system import (
    WindPowerPlantSystem as WindPowerPlantSystem,
)



"""This module contains the state model of the chp system."""

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class CHPLPGSystemState(ModelState):
    """captures the state of the chp system

    The references are updated during each step.
    """

    def __init__(self, init_vals: ModelInitVals):
        super().__init__(init_vals)

        self.p_th_kw = 0.0
        """Current thermal power of the chp in [kW]."""

        self.storage_t_c = 0.0
        """Current temperature of the heat storage in [°C]."""

        self.lubricant_l = 0.0
        """Among of lubricant remaining in [l]."""

        self.e_th_demand_kwh = 0.0
        """Current heat demand of the household in [kW]"""



"""This module contains a CHP model that is used by the biogas plant."""

from copy import deepcopy
from datetime import timedelta

import numpy as np

from pysimmods.base.generator import Generator
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.generator.chpcngsim.config import CHPCNGConfig
from pysimmods.generator.chpcngsim.inputs import CHPCNGInputs
from pysimmods.generator.chpcngsim.state import CHPCNGState


class CHPCNG(Generator[CHPCNGConfig, CHPCNGState, CHPCNGInputs]):
    """A combined heat and power unit model.

    The CHP is fueled with Compressed Natural Gas (CNG).
    The main purpose for this model is to be used by the biogas plant.

    Attributes
    ----------
    config : CHPLPGConfig
    state : CHPLPGState
    inputs : CHPLPGInputs

    """

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = CHPCNGConfig(params)
        self.state = CHPCNGState(inits)
        self.inputs = CHPCNGInputs()

    def step(self, pretend=False) -> CHPCNGState:
        """Perform a simulation step."""
        next_state = deepcopy(self.state)
        next_state.p_kw = self._get_setpoint()

        self._check_constraints(next_state)

        if next_state.p_kw != 0:
            # Unit is / switches on
            next_state.active_s_per_day += self.inputs.step_size
            next_state.inactive_s = 0
            if self.state.p_kw != 0:
                # Unit was on in the last step
                if self.state.p_kw != next_state.p_kw:
                    # Power level changed
                    next_state.active_s = self.inputs.step_size
                else:
                    next_state.active_s += self.inputs.step_size
            else:
                next_state.restarts += 1
        else:
            # Unit is / switches off
            next_state.inactive_s += self.inputs.step_size
            next_state.inactive_s_per_day += (
                self.inputs.step_size % 86_400
            )  # Works correctly?

        next_state.p_th_kw = self._gas2thermal_power(
            next_state.p_kw, next_state.gas_cons_m3
        )

        # Update the time
        # TODO: Handle odd step sizes.
        next_dt = self.inputs.now_dt + timedelta(seconds=self.inputs.step_size)
        if next_dt.day != self.inputs.now_dt.day:
            # Start a new day
            next_state.restarts = 0
            next_state.active_s_per_day = 0
            next_state.inactive_s_per_day = 0

        if not pretend:
            self.state = next_state

        return next_state

    def _get_setpoint(self):
        setpoint = self.inputs.p_set_kw

        if setpoint is None:
            hour = self.inputs.now_dt.hour
            default = self.config.default_schedule[hour]
            setpoint = self.config.p_min_kw + (
                (self.config.p_max_kw - self.config.p_min_kw) * default / 100
            )

        return abs(setpoint)

    def _check_constraints(self, next_state):
        if not self.inputs.gas_critical:
            self._check_daily_incative_time(next_state)
            self._check_restarts(next_state)
            self._check_daily_active_time(next_state)

        self._check_gas_demand(next_state)

    def _check_daily_incative_time(self, next_state):
        # Only check if this a constraint
        if self.config.inactive_max_s_per_day > 0:
            # Unit should be switched off in the next state
            if next_state.p_kw == 0:
                # Total inactive time after this step
                inactive_next = (
                    next_state.inactive_s_per_day + self.inputs.step_size
                )

                # No more inactivity allowed for today
                if inactive_next > self.config.inactive_max_s_per_day:
                    next_state.p_kw = self.config.pn_min_kw

    def _check_restarts(self, next_state):
        # Only check if this is a constraint
        if self.config.restarts_per_day > 0:
            # Unit was off and should switch on
            if self.state.p_kw == 0 and next_state.p_kw != 0:
                # Another restart is invalid
                if next_state.restarts >= self.config.restarts_per_day:
                    next_state.p_kw = 0

    def _check_daily_active_time(self, next_state):
        # Only check if this is a constraint
        if self.config.active_max_s_per_day > 0:
            # Unit should switch/stay on in the next step
            if next_state.p_kw != 0:
                # Total active time after this step
                active_next = (
                    next_state.active_s_per_day + self.inputs.step_size
                )
                # Unit cannot stay switched on
                if active_next > self.config.active_max_s_per_day:
                    next_state.p_kw = 0

    def _check_gas_demand(self, next_state):
        # Unit should operate on next step
        if next_state.p_kw != 0:
            gas_cons = self._gas_demand(next_state.p_kw)

            # Enough gas for a whole step
            if gas_cons < self.inputs.gas_in_m3:
                next_state.gas_cons_m3 = gas_cons
            else:
                # Do not consider running for half a step
                next_state.p_kw = 0
                next_state.gas_cons_m3 = 0
        else:
            next_state.gas_cons_m3 = 0

    def _gas_demand(self, p_kw):
        e_kwh = abs(p_kw) * self.inputs.step_size / 3_600

        # Get best matching power level
        idx = np.argmin(abs(np.array(self.config.pn_stages_kw) - p_kw))
        # Get eta for p_kw
        eta = self.config.eta_stages_percent[idx] / 100

        gas = e_kwh / (
            eta
            * self.config.e_ch4_kwh
            * self.config.ch4_concentration_percent
            / 100
        )
        return gas

    def _gas2thermal_power(self, p_kw, gas):
        idx = np.argmin(abs(np.array(self.config.pn_stages_kw) - p_kw))
        eta = self.config.eta_th_stages_percent[idx] / 100
        v_methane = (
            self.config.e_ch4_kwh
            * self.config.ch4_concentration_percent
            * gas
            / 100
        )
        p_th_kw = eta * v_methane * 3_600 / self.inputs.step_size

        return p_th_kw



import os

import pandas as pd
from midas.util import report_util

from .power import analyze_power


def analyze(
    name: str,
    data: pd.DataFrame,
    output_folder: str,
    start: int,
    end: int,
    step_size: int,
    full: bool,
):
    all_der_columns = [col for col in data.columns if "Pysimmods" in col]
    # der_sim_keys = [
    #     sim_key for sim_key in data.keys() if "Pysimmods" in sim_key
    # ]
    der_sim_keys = list(
        dict.fromkeys([c.split(".", 1)[0] for c in all_der_columns])
    )

    for sim_key in der_sim_keys:
        sim_cols = [col for col in data.columns if sim_key in col]
        der_data = data[sim_cols]
        if start > 0:
            der_data = der_data.iloc[start:]
        if end > 0:
            der_data = der_data.iloc[:end]

        analyze_der(
            der_data,
            step_size,
            f"{name}-{sim_key.replace('/', '')}",
            output_folder,
            full,
        )


def analyze_der(data, step_size, name, output_path, full_report):
    plot_path = os.path.join(
        output_path, name.split("-", 1)[1].replace("__", "_")
    )
    os.makedirs(plot_path, exist_ok=True)

    ef = step_size / 3_600
    report_content = []

    gens = ["Photovoltaic", "CHP", "Biogas", "DieselGenerator", "WindTurbine"]
    loads = ["HVAC"]
    bufs = ["Battery"]

    model_totals = {}
    for model in gens + loads + bufs:
        mod_data = data[[col for col in data.columns if model in col]]
        if not mod_data.empty:
            model_totals[model] = analyze_power(
                mod_data,
                step_size,
                report_content,
                plot_path,
                full_report,
                {"name": name, "topic": model, "total_name": f"{model}s"},
            )

    report_path = os.path.join(output_path, f"{name}_report.md")
    report_file = open(report_path, "w")

    total_gen_p = 0
    total_gen_q = 0
    total_load_p = 0
    total_load_q = 0
    for model, totals in model_totals.items():
        if model in gens:
            total_gen_p += totals[0].sum()
            total_gen_q += totals[1].sum()
        if model in loads:
            total_load_p += totals[0].sum()
            total_load_q += totals[1].sum()
        if model in bufs:
            total_gen_p -= totals[0][totals[0] < 0].sum()
            total_gen_q -= totals[1][totals[1] < 0].sum()
            total_load_p += totals[0][totals[0] > 0].sum()
            total_load_q += totals[1][totals[1] > 0].sum()

    report_file.write(
        f"# Analysis of {name}\n\n## Summary\n\n"
        f"* total active generation: {total_gen_p * ef:.2f} MWh\n"
        f"* total reactive generation: {total_gen_q * ef:.2f} MVArh\n"
        f"* total active consumption: {total_load_p * ef:.2f} MWh\n"
        f"* total reactive consumption: {total_load_q * ef:.2f} MVArh\n\n"
    )

    for line in report_content:
        report_file.write(f"{line}\n")
    report_file.close()

    report_util.convert_markdown(report_path)



"""This module contains the input model for PV plant."""

from typing import Optional

from pysimmods.base.inputs import ModelInputs

CELSIUS_TO_KELVIN: float = 273.15
KELVIN_TO_CELSIUS: float = -1 * CELSIUS_TO_KELVIN
HPA_TO_PA: float = 100.0
PA_TO_HPA: float = 1 / HPA_TO_PA


class WindPowerPlantInputs(ModelInputs):
    """Input variables of Wind plant model.

    See :class:`pysimmods.model.inputs.ModelInputs` for additional
    information. This class has no inputs itself. Instead, each
    of the values is to be provided before each step.

    Attributes
    ----------
    t_air_deg_celsius:
        Air temperature in [°C].
    t_air_deg_kelvin:
        Air temperature in [°K].
    wind_v_m_per_s:
        Wind speed in [m/s].
    wind_dir_deg:
        Wind direction in degrees (currently not used).
    air_pressure_hpa:
        Air pressure in [hPa]
    pressure_pa:
        Air pressure in [Pa].

    """

    def __init__(self):
        super().__init__()

        self.t_air_deg_celsius: float = 0.0
        self.wind_v_m_per_s: float = 0.0
        self.wind_dir_deg: Optional[float] = None
        self.air_pressure_hpa: float = 0.0

    @property
    def pressure_pa(self):
        return self.air_pressure_hpa * HPA_TO_PA

    @pressure_pa.setter
    def pressure_pa(self, value):
        if value < 50000:
            raise ValueError("pressure value error")
        self.air_pressure_hpa = value * PA_TO_HPA

    @property
    def t_air_deg_kelvin(self):
        return self.t_air_deg_celsius + CELSIUS_TO_KELVIN

    @t_air_deg_kelvin.setter
    def t_air_deg_kelvin(self, value):
        if value < 0:
            raise ValueError("temperature value error")
        self.t_air_deg_celsius = value + KELVIN_TO_CELSIUS



from pysimmods.generator.windsystemsim.inputs import ModelInputs


class HeatStorageInputs(ModelInputs):
    """Captures the inputs of the heat storage"""

    def __init__(self):

        self.e_th_prod_kwh: float = 0.0
        self.e_th_demand_kwh: float = 0.0


"""This module contains the state model for pv."""

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class WindPowerPlantState(ModelState):
    """State parameters of Wind model.

    See :class:`pysimmods.model.state.ModelState` for additional
    information.

    Parameters
    ----------
    inits : dict
        Contains the initial configuration of this wind plant. See
        attributes section for specific to the wind plant.

    Attributes
    ----------
    wind_hub_v_m_per_s: float
        Calculated wind speed at hub height in [m/s]
    t_air_hub_deg_kelvin: float
        Calculated air temperature at hub height in [°K].
    air_density_hub_kg_per_m3: float
        Calculated air density at hub height in [kg/m^3].

    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.wind_hub_v_m_per_s: float = 0.0
        self.t_air_hub_deg_kelvin: float = 0.0
        self.air_density_hub_kg_per_m3: float = 0.0



"""This module contains the config model for the chp."""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class CHPCNGConfig(ModelConfig):
    """A CHP CNG configuration parameter class.

    Parameters
    ----------
    params: dict
        Contains the configuration for the CHP CNG. See attribute
        section for more information about the parameters.
        Attributes marked with *(Input)* can be provided with this
        dict, others are calculated from those.

    Attributes
    ----------
    pn_stages_kw: list[float]
        (Input) Possible eletrical power stages in [kW]. Power level 0
        (unit off) should not be included.
    p_max_kw: float
        The maximum power output of this unit in [kW].
    p_min_kw: float
        The minimum power output of this unit in [kW], while it is
        operating (exluding 0).
    eta_stages_percent: list[float]
        (Input) The electrical efficiency at each power stage in [%]
    eta_th_stages_percent: list[float]
        (Input) The thermal efficiency at each power stage in [%]
    restarts_per_day: int
        (Input) Allowed number of restarts per day.
    active_min_s: int
        (Input) Minimum active time of this unit before it is allowed
        to shut down again in [s].
    active_max_s_per_day: int
        (Input) Maximum cumulated active time per day in [s]. Unit will
        shut down after that time.
    inactive_min_s: int
        (Input) Minimum inactive time until this unit is allowed to
        start again in [s].
    inactive_max_s_per_day: int
        (Input) Maximum cumulated inactive time per day in [s]. Unit
        will try to stay on when this time is reached.
    e_ch4_kwh: float
        Energy of methane gas in [kWh]. Constant value.
    ch4_concentration_percent: float
        (Input) Concentration of methane gas in [%].
    """

    def __init__(self, params: ModelParams) -> None:
        super().__init__(params)

        self.pn_stages_kw: list[float] = cast(
            "list[float]", params["pn_stages_kw"]
        )
        self.pn_stages_kw = [abs(stage) for stage in self.pn_stages_kw]
        self.p_max_kw: float = max(self.pn_stages_kw)
        self.p_min_kw: float = min(self.pn_stages_kw)

        self.eta_stages_percent: list[float] = cast(
            "list[float]", params["eta_stages_percent"]
        )
        self.eta_th_stages_percent: list[float] = cast(
            "list[float]", params["eta_th_stages_percent"]
        )

        self.restarts_per_day: int = cast("int", params["restarts_per_day"])
        self.active_min_s: int = cast("int", params["active_min_s"])
        self.active_max_s_per_day: int = cast(
            "int", params["active_max_s_per_day"]
        )
        self.inactive_min_s: int = cast("int", params["inactive_min_s"])
        self.inactive_max_s_per_day: int = cast(
            "int", params["inactive_max_s_per_day"]
        )

        self.e_ch4_kwh = 9.94
        self.ch4_concentration_percent: float = cast(
            "float", params.get("ch4_concentration_percent", 50.302)
        )

        self.default_schedule: list[float] = [50.0] * 24



M_AIR_KG_PER_MOL: float = 0.029
ACCELERATION_GRAVITY_M_PER_S2: float = 9.81
GAS_CONSTANT_J_PER_MOL_K: float = 8.314
GAS_CONSTANT_J_PER_KG_K: float = 287.058
T_0_DEGREE_KELVIN: float = 288.15
P_0_HPA: float = 1013.25
P_GRADIENT_HPA_PER_M: float = 1 / 8
AIR_DENSITY_KG_PER_M3: float = 1.225


def barometric(
    pressure_hpa: float,
    t_air_hub_deg_kelvin: float,
    data_height_m: float,
    target_height_m: float,
) -> float:
    """Calculate pressure at altitude h using barometric height equation.

    The barometric height equation to calculate the air density at hub
    height using pressure and temperature at hub height.
    Uses an air density constant of 1.225 [kg/m^3] and the default
    pressure of 1013.25 [hPa].

    Parameters
    ----------
    pressure_hpa: float
        The air pressure that was measured in [hPa].
    t_air_hub_deg_kelvin: float
        The air temperature at altitude in [°K].
    data_height_m: float
        The height in which the data was measured in [m].
    target_height_m: float
        The height for which the pressure should be calculated in [m].

    Returns
    -------
    float
        The air pressure at altitude `target_height_m`.

    """

    return (
        pressure_correction(pressure_hpa, data_height_m, target_height_m)
        * AIR_DENSITY_KG_PER_M3
        * T_0_DEGREE_KELVIN
        / (P_0_HPA * t_air_hub_deg_kelvin)
    )


def ideal_gas(
    pressure_hpa: float,
    t_air_hub_deg_kelvin: float,
    data_height_m: float,
    target_height_m: float,
) -> float:
    """Calculate air pressure using ideal gas method.

    The barometric height equation to calculate the air density at hub
    height using pressure and Temperature at hub height
    gas constant of dry air (287.058 J/(kg*K))

    Returns
    -------
    float
        Air density at `target_height_m` in [hPa].
    """

    return (
        pressure_correction(pressure_hpa, data_height_m, target_height_m)
        / (GAS_CONSTANT_J_PER_KG_K * t_air_hub_deg_kelvin)
        * 100
    )


def pressure_correction(
    pressure_hpa: float, data_height_m: float, target_height_m: float
) -> float:
    """Calculate pressure correction based on a gradient.

    The gradient is assumed to be 1/8 hPa/m.

    Parameters
    ----------
    pressure_hpa: float
        The measured air pressure in [hPa].
    data_height_m: float
        Height in which the data was measured in [m].
    target_height_m: float
        Height for which the pressure should be calculated in [m].

    """
    return (
        pressure_hpa - (target_height_m - data_height_m) * P_GRADIENT_HPA_PER_M
    )



"""This module contains the config information for the heat demand
model.

"""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class HeatDemandConfig(ModelConfig):
    """Captures the configuration variables of the heat demand model."""

    def __init__(self, params: ModelParams):
        super().__init__(params)
        self.const_a: float = cast("float", params.get("A"))
        self.const_b: float = cast("float", params.get("B"))
        self.const_c: float = cast("float", params.get("C"))
        self.const_d: float = cast("float", params.get("D"))
        self.const_v_0: float = cast("float", params.get("V_0"))
        self.const_m_h: float = cast("float", params.get("M_H"))
        self.const_b_h: float = cast("float", params.get("B_H"))
        self.const_m_w: float = cast("float", params.get("M_W"))
        self.const_b_w: float = cast("float", params.get("B_W"))

        self.load_profile: float = cast(
            "list[float]", params.get("load_profile")
        )
        self.consumer_const: float = cast(
            "float", params.get("consumer_constant")
        )
        self.weekday_const: float = cast(
            "list[float]", params.get("weekday_constants")
        )

        self.degree_of_efficiency = 0.93
